from typing import List

from fastapi import APIRouter, Depends
from sqlmodel import select

from murfey.server.api.auth import validate_token
from murfey.server.murfey_db import murfey_db
from murfey.util.db import MagnificationLookup

router = APIRouter(
    prefix="/mag_table",
    dependencies=[Depends(validate_token)],
    tags=["Magnification Table"],
)


@router.get("/mag_table/")
def get_mag_table(db=murfey_db) -> List[MagnificationLookup]:
    return db.exec(select(MagnificationLookup)).all()


@router.post("/mag_table/")
def add_to_mag_table(rows: List[MagnificationLookup], db=murfey_db):
    for r in rows:
        db.add(r)
    db.commit()


@router.delete("/mag_table/{mag}")
def remove_mag_table_row(mag: int, db=murfey_db):
    row = db.exec(
        select(MagnificationLookup).where(MagnificationLookup.magnification == mag)
    ).one()
    db.delete(row)
    db.commit()



from unittest import mock

from sqlmodel import Session, select

from murfey.util.db import DataCollectionGroup, SearchMap, TiltSeries
from murfey.util.models import BatchPositionParameters, SearchMapParameters
from murfey.workflows.tomo import tomo_metadata
from tests.conftest import ExampleVisit


@mock.patch("murfey.workflows.tomo.tomo_metadata._transport_object")
def test_register_search_map_update_with_dimensions(
    mock_transport, murfey_db_session: Session
):
    """Test the updating of an existing search map, without enough to find location"""
    # Create a search map to update
    search_map = SearchMap(
        id=1,
        name="SearchMap_1",
        session_id=ExampleVisit.murfey_session_id,
        tag="session_tag",
        x_stage_position=0.1,
        y_stage_position=0.2,
    )
    murfey_db_session.add(search_map)
    murfey_db_session.commit()

    # Make sure DCG is present with a pixel size
    dcg = DataCollectionGroup(
        id=1,
        session_id=ExampleVisit.murfey_session_id,
        tag="session_tag",
        atlas_id=90,
        atlas_pixel_size=1e-5,
    )
    murfey_db_session.add(dcg)
    murfey_db_session.commit()

    # Parameters to update with
    new_parameters = SearchMapParameters(
        tag="session_tag",
        width=2000,
        height=4000,
    )

    # Run the registration
    tomo_metadata.register_search_map_in_database(
        ExampleVisit.murfey_session_id, "SearchMap_1", new_parameters, murfey_db_session
    )

    # Check this would have updated ispyb
    mock_transport.do_update_search_map.assert_called_with(1, new_parameters)

    # Confirm the database was updated
    sm_final_parameters = murfey_db_session.exec(select(SearchMap)).one()
    assert sm_final_parameters.width == new_parameters.width
    assert sm_final_parameters.height == new_parameters.height
    assert sm_final_parameters.x_stage_position == 0.1
    assert sm_final_parameters.y_stage_position == 0.2
    assert sm_final_parameters.x_location is None


@mock.patch("murfey.workflows.tomo.tomo_metadata._transport_object")
def test_register_search_map_update_with_all_parameters(
    mock_transport, murfey_db_session: Session
):
    """Test the updating of an existing search map with all required parameters"""
    # Create a search map to update
    search_map = SearchMap(
        id=1,
        name="SearchMap_1",
        session_id=ExampleVisit.murfey_session_id,
        tag="session_tag",
        x_stage_position=0.1,
        y_stage_position=0.2,
        width=2000,
        height=4000,
    )
    murfey_db_session.add(search_map)
    murfey_db_session.commit()

    # Make sure DCG is present with a pixel size
    dcg = DataCollectionGroup(
        id=1,
        session_id=ExampleVisit.murfey_session_id,
        tag="session_tag",
        atlas_id=90,
        atlas_pixel_size=1e-5,
    )
    murfey_db_session.add(dcg)
    murfey_db_session.commit()

    # Parameters to update with
    new_parameters = SearchMapParameters(
        tag="session_tag",
        x_stage_position=0.3,
        y_stage_position=0.4,
        pixel_size=1e-7,
        image="path/to/image",
        binning=1,
        reference_matrix={"m11": 1.01, "m12": 0.01, "m21": 0.02, "m22": 1.02},
        stage_correction={"m11": 0.99, "m12": -0.01, "m21": -0.02, "m22": 0.98},
        image_shift_correction={"m11": 1.03, "m12": 0.03, "m21": -0.03, "m22": 0.97},
    )

    # Run the registration
    tomo_metadata.register_search_map_in_database(
        ExampleVisit.murfey_session_id, "SearchMap_1", new_parameters, murfey_db_session
    )

    # Confirm the database was updated
    sm_final_parameters = murfey_db_session.exec(select(SearchMap)).one()
    assert sm_final_parameters.width == 2000
    assert sm_final_parameters.height == 4000
    assert sm_final_parameters.x_stage_position == 0.3
    assert sm_final_parameters.y_stage_position == 0.4
    assert sm_final_parameters.pixel_size == 1e-7
    assert sm_final_parameters.image == "path/to/image"
    assert sm_final_parameters.binning == 1
    assert sm_final_parameters.reference_matrix_m11 == 1.01
    assert sm_final_parameters.reference_matrix_m12 == 0.01
    assert sm_final_parameters.reference_matrix_m21 == 0.02
    assert sm_final_parameters.reference_matrix_m22 == 1.02
    assert sm_final_parameters.stage_correction_m11 == 0.99
    assert sm_final_parameters.stage_correction_m12 == -0.01
    assert sm_final_parameters.stage_correction_m21 == -0.02
    assert sm_final_parameters.stage_correction_m22 == 0.98
    assert sm_final_parameters.image_shift_correction_m11 == 1.03
    assert sm_final_parameters.image_shift_correction_m12 == 0.03
    assert sm_final_parameters.image_shift_correction_m21 == -0.03
    assert sm_final_parameters.image_shift_correction_m22 == 0.97

    # These two should have been updated, but what that update should be is messy
    assert sm_final_parameters.x_location is not None
    assert sm_final_parameters.y_location is not None

    # Check this would have updated ispyb
    mock_transport.do_update_search_map.assert_called_with(1, new_parameters)
    new_parameters.x_location = sm_final_parameters.x_location
    new_parameters.y_location = sm_final_parameters.y_location
    new_parameters.height_on_atlas = 40
    new_parameters.width_on_atlas = 20
    mock_transport.do_update_search_map.assert_called_with(1, new_parameters)


@mock.patch("murfey.workflows.tomo.tomo_metadata._transport_object")
def test_register_search_map_insert_with_ispyb(
    mock_transport, murfey_db_session: Session, tmp_path
):
    """Insert a new search map"""
    # Create a data collection group for lookups
    dcg = DataCollectionGroup(
        id=1,
        session_id=ExampleVisit.murfey_session_id,
        tag="session_tag",
        atlas_id=90,
        atlas_pixel_size=1e-5,
    )
    murfey_db_session.add(dcg)
    murfey_db_session.commit()

    # Set the ispyb return
    mock_transport.do_insert_search_map.return_value = {
        "return_value": 1,
        "success": True,
    }

    # Parameters to update with
    new_parameters = SearchMapParameters(
        tag="session_tag",
        x_stage_position=1.3,
        y_stage_position=1.4,
        pixel_size=1.02,
    )

    # Run the registration
    tomo_metadata.register_search_map_in_database(
        ExampleVisit.murfey_session_id, "SearchMap_1", new_parameters, murfey_db_session
    )

    # Check this would have updated ispyb
    mock_transport.do_insert_search_map.assert_called_with(90, new_parameters)

    # Confirm the database entry was made
    sm_final_parameters = murfey_db_session.exec(select(SearchMap)).one()
    assert sm_final_parameters.id == 1
    assert sm_final_parameters.name == "SearchMap_1"
    assert sm_final_parameters.session_id == ExampleVisit.murfey_session_id
    assert sm_final_parameters.tag == "session_tag"
    assert sm_final_parameters.x_stage_position == 1.3
    assert sm_final_parameters.y_stage_position == 1.4
    assert sm_final_parameters.pixel_size == 1.02
    assert sm_final_parameters.x_location is None


def test_register_batch_position_update(murfey_db_session: Session):
    """Test the updating of an existing tilt series with batch positions"""
    # Make sure search map is present
    search_map = SearchMap(
        id=1,
        name="SearchMap_1",
        session_id=ExampleVisit.murfey_session_id,
        tag="session_tag",
        x_stage_position=1,
        y_stage_position=2,
        pixel_size=0.01,
        reference_matrix_m11=1,
        reference_matrix_m12=0,
        reference_matrix_m21=0,
        reference_matrix_m22=1,
        stage_correction_m11=1,
        stage_correction_m12=0,
        stage_correction_m21=0,
        stage_correction_m22=1,
        image_shift_correction_m11=1,
        image_shift_correction_m12=0,
        image_shift_correction_m21=0,
        image_shift_correction_m22=1,
        height=4000,
        width=2000,
    )
    murfey_db_session.add(search_map)
    murfey_db_session.commit()

    # Create a tilt series to update
    tilt_series = TiltSeries(
        tag="Position_1",
        rsync_source="session_tag",
        session_id=ExampleVisit.murfey_session_id,
        search_map_id=1,
    )
    murfey_db_session.add(tilt_series)
    murfey_db_session.commit()

    # Parameters to update with
    new_parameters = BatchPositionParameters(
        tag="session_tag",
        x_stage_position=0.1,
        y_stage_position=0.2,
        x_beamshift=0.3,
        y_beamshift=0.4,
        search_map_name="SearchMap_1",
    )

    # Run the registration
    tomo_metadata.register_batch_position_in_database(
        ExampleVisit.murfey_session_id, "Position_1", new_parameters, murfey_db_session
    )

    # These two should have been updated, values are known as used identity matrices
    bp_final_parameters = murfey_db_session.exec(select(TiltSeries)).one()
    assert bp_final_parameters.x_location == 880 * 512 / 2000
    assert bp_final_parameters.y_location == (4000 - 1780) * 512 / 2000


def test_register_batch_position_update_skip(murfey_db_session: Session):
    """Test the updating of an existing batch position, skipped as already done"""
    # Make sure search map is present
    search_map = SearchMap(
        id=1,
        name="SearchMap_1",
        session_id=ExampleVisit.murfey_session_id,
        tag="session_tag",
        x_stage_position=1,
        y_stage_position=2,
        pixel_size=0.01,
        reference_matrix_m11=1,
        reference_matrix_m12=0,
        reference_matrix_m21=0,
        reference_matrix_m22=1,
        stage_correction_m11=1,
        stage_correction_m12=0,
        stage_correction_m21=0,
        stage_correction_m22=1,
        image_shift_correction_m11=1,
        image_shift_correction_m12=0,
        image_shift_correction_m21=0,
        image_shift_correction_m22=1,
        height=4000,
        width=2000,
    )
    murfey_db_session.add(search_map)
    murfey_db_session.commit()

    # Create a tilt series to update
    tilt_series = TiltSeries(
        tag="Position_1",
        rsync_source="session_tag",
        session_id=ExampleVisit.murfey_session_id,
        search_map_id=1,
        x_location=100,
        y_location=200,
    )
    murfey_db_session.add(tilt_series)
    murfey_db_session.commit()

    # Parameters to update with
    new_parameters = BatchPositionParameters(
        tag="session_tag",
        x_stage_position=0.1,
        y_stage_position=0.2,
        x_beamshift=0.3,
        y_beamshift=0.4,
        search_map_name="SearchMap_1",
    )

    # Run the registration
    tomo_metadata.register_batch_position_in_database(
        ExampleVisit.murfey_session_id, "Position_1", new_parameters, murfey_db_session
    )

    # These two should have been updated, values are known as used identity matrices
    bp_final_parameters = murfey_db_session.exec(select(TiltSeries)).one()
    assert bp_final_parameters.x_location == 100
    assert bp_final_parameters.y_location == 200


def test_register_batch_position_new(murfey_db_session: Session):
    """Test the registration of a new tilt series with batch positions"""
    # Make sure search map is present
    search_map = SearchMap(
        id=1,
        name="SearchMap_1",
        session_id=ExampleVisit.murfey_session_id,
        tag="session_tag",
        x_stage_position=1,
        y_stage_position=2,
        pixel_size=0.01,
        reference_matrix_m11=1,
        reference_matrix_m12=0,
        reference_matrix_m21=0,
        reference_matrix_m22=1,
        stage_correction_m11=1,
        stage_correction_m12=0,
        stage_correction_m21=0,
        stage_correction_m22=1,
        image_shift_correction_m11=1,
        image_shift_correction_m12=0,
        image_shift_correction_m21=0,
        image_shift_correction_m22=1,
        height=4000,
        width=2000,
    )
    murfey_db_session.add(search_map)
    murfey_db_session.commit()

    # Parameters to update with
    new_parameters = BatchPositionParameters(
        tag="session_tag",
        x_stage_position=0.1,
        y_stage_position=0.2,
        x_beamshift=0.3,
        y_beamshift=0.4,
        search_map_name="SearchMap_1",
    )

    # Run the registration
    tomo_metadata.register_batch_position_in_database(
        ExampleVisit.murfey_session_id, "Position_1", new_parameters, murfey_db_session
    )

    # These two should have been updated, values are known as used identity matrices
    bp_final_parameters = murfey_db_session.exec(select(TiltSeries)).one()
    assert bp_final_parameters.x_location == 880 * 512 / 2000
    assert bp_final_parameters.y_location == (4000 - 1780) * 512 / 2000



from __future__ import annotations

import logging

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware

from murfey.instrument_server.api import router

log = logging.getLogger("murfey.instrument_server.main")

app = FastAPI(title="Murfey instrument server", debug=True)

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

app.include_router(router)



"""
Script to allow Murfey to submit the TIFF-to-stack job to the cluster.
The recipe referred to here is stored on GitLab.
"""

from logging import getLogger
from pathlib import Path
from typing import Optional

logger = getLogger("murfey.workflows.clem.process_raw_tiffs")

try:
    from murfey.server.ispyb import TransportManager  # Session
except AttributeError:
    pass  # Ignore if ISPyB credentials environment variable not set


def zocalo_cluster_request(
    tiff_list: list[Path],
    root_folder: str,
    session_id: int,
    instrument_name: str,
    metadata: Optional[Path] = None,
    messenger: Optional[TransportManager] = None,
):
    if messenger:
        # Construct path to session directory
        path_parts = list(
            (tiff_list[0].parent / (tiff_list[0].stem.split("--")[0])).parts
        )
        # Replace leading "/" in Unix paths
        path_parts[0] = "" if path_parts[0] == "/" else path_parts[0]
        try:
            # Find the position of the root folder in the list
            root_index = [p.lower() for p in path_parts].index(root_folder.lower())
        except ValueError:
            raise Exception(
                f"Unable to find the root folder {root_folder!r} in the file path {tiff_list[0]!r}"
            )
        # Construct the session and job name
        session_dir = "/".join(path_parts[:root_index])
        job_name = "--".join(
            [p.replace(" ", "_") if " " in p else p for p in path_parts][
                root_index + 1 :
            ]
        )

        # If no metadata file provided, generate path to one
        if metadata is None:
            series_name = tiff_list[0].stem.split("--")[0]
            metadata = tiff_list[0].parent / "Metadata" / (series_name + ".xlif")

        # Load machine config to get the feedback queue
        feedback_queue: str = messenger.feedback_queue

        # Construct recipe and submit it for processing
        recipe = {
            "recipes": ["clem-tiff-to-stack"],
            "parameters": {
                # Job parameters
                "tiff_list": "null",
                "tiff_file": f"{str(tiff_list[0])}",
                "root_folder": root_folder,
                "metadata": f"{str(metadata)}",
                # Other recipe parameters
                "session_dir": f"{str(session_dir)}",
                "session_id": session_id,
                "job_name": job_name,
                "feedback_queue": feedback_queue,
            },
        }
        logger.debug(
            f"Submitting TIFF processing request to {messenger.feedback_queue!r} "
            "with the following recipe: \n"
            f"{recipe}"
        )
        messenger.send(
            queue="processing_recipe",
            message=recipe,
            new_connection=True,
        )
    else:
        raise Exception("Unable to find transport manager")



from __future__ import annotations

import json
import logging

from rich.logging import RichHandler

from murfey.client.tui.screens import LogBook

logger = logging.getLogger("murfey.client.customlogging")


class CustomHandler(logging.Handler):
    def __init__(self, callback):
        """Set up a handler instance, record the callback function."""
        super().__init__()
        self._callback = callback

    def prepare(self, record):
        self.format(record)
        record_dict = record.__dict__
        record_dict["type"] = "log"
        try:
            return json.dumps(record_dict)
        except TypeError:
            return json.dumps({str(k): str(v) for k, v in record_dict.items()})

    def emit(self, record):
        try:
            self._callback(self.prepare(record))
        except Exception:
            self.handleError(record)


class DirectableRichHandler(RichHandler):
    def __init__(self, text_log: LogBook | None = None, **kwargs):
        super().__init__(**kwargs)
        self.text_log = text_log
        self.redirect = False
        self._last_time = None

    def emit(self, record):
        try:
            if self.text_log:
                message = self.format(record)
                message_renderable = self.render_message(record, message)
                log_renderable = self.render(
                    record=record, traceback=None, message_renderable=message_renderable
                )
                self.text_log.post_message(self.text_log.Log(log_renderable))
        except Exception:
            self.handleError(record)



import argparse
import os
import time
from pathlib import Path

import sqlalchemy
import workflows
import zocalo
from sqlmodel import Session, create_engine, select

from murfey.server.ispyb import TransportManager
from murfey.server.murfey_db import url
from murfey.util.config import get_machine_config, get_microscope, get_security_config
from murfey.util.db import (
    AutoProcProgram,
    ClientEnvironment,
    DataCollection,
    DataCollectionGroup,
    Movie,
    ProcessingJob,
    SPAFeedbackParameters,
    SPARelionParameters,
)
from murfey.util.processing_params import default_spa_parameters


def run():
    parser = argparse.ArgumentParser(
        description="Inject movies for SPA processing from Murfey database movie store"
    )

    parser.add_argument(
        "--tag",
        dest="tag",
        type=str,
        required=True,
        help="Tag from Murfey database Movie table",
    )
    parser.add_argument(
        "-s",
        "--session-id",
        dest="session_id",
        required=True,
        type=int,
        help="Murfey session ID",
    )
    parser.add_argument(
        "--max-image-number",
        dest="max_image_number",
        default=-1,
        type=int,
        help="Upper end of image number range",
    )
    parser.add_argument(
        "--min-image-number",
        dest="min_image_number",
        default=1,
        type=int,
        help="Lower end of image number range",
    )
    parser.add_argument(
        "--injection-delay",
        dest="injection_delay",
        default=0.5,
        type=float,
        help="Time spacing between processing requests in seconds",
    )
    parser.add_argument(
        "--ignore-processing-status",
        dest="check_preproc",
        default=True,
        action="store_false",
        help="Inject movies even when they are already tagged as preprocessed",
    )
    parser.add_argument(
        "-m",
        "--microscope",
        dest="microscope",
        type=str,
        default="",
        help="Microscope as specified in the Murfey machine configuration",
    )
    parser.add_argument(
        "--eer-fractionation-file",
        dest="eer_fractionation_file",
        default=None,
        help="Path to EER fractionation file if relevant",
    )

    zc = zocalo.configuration.from_file()
    zc.activate()
    zc.add_command_line_options(parser)
    workflows.transport.add_command_line_options(parser, transport_argument=True)

    args = parser.parse_args()
    if args.microscope:
        os.environ["BEAMLINE"] = args.microscope

    machine_config = get_machine_config()
    security_config = get_security_config()
    _url = url(machine_config)
    engine = create_engine(_url)
    murfey_db = Session(engine)

    _transport_object = TransportManager(args.transport)
    _transport_object.feedback_queue = security_config.feedback_queue

    query = (
        select(Movie)
        .where(Movie.tag == args.tag)
        .where(Movie.image_number >= args.min_image_number)
    )
    if args.max_image_number > 0:
        query = query.where(Movie.image_number <= args.max_image_number)
    if args.check_preproc:
        query = query.where(Movie.preprocessed.is_(False))
    movies = murfey_db.exec(query).all()

    visit_name = (
        murfey_db.exec(
            select(ClientEnvironment).where(
                ClientEnvironment.session_id == args.session_id
            )
        )
        .one()
        .visit
    )

    try:
        collected_ids = murfey_db.exec(
            select(DataCollectionGroup, DataCollection, ProcessingJob, AutoProcProgram)
            .where(DataCollectionGroup.session_id == args.session_id)
            .where(DataCollectionGroup.tag == args.tag)
            .where(DataCollection.dcg_id == DataCollectionGroup.id)
            .where(ProcessingJob.dc_id == DataCollection.id)
            .where(AutoProcProgram.pj_id == ProcessingJob.id)
            .where(ProcessingJob.recipe == "em-spa-preprocess")
        ).one()
        params = murfey_db.exec(
            select(SPARelionParameters, SPAFeedbackParameters)
            .where(SPARelionParameters.pj_id == collected_ids[2].id)
            .where(SPAFeedbackParameters.pj_id == SPARelionParameters.pj_id)
        ).one()
        proc_params: dict | None = dict(params[0])
        feedback_params = params[1]
        if feedback_params.picker_murfey_id is None:
            raise ValueError("No ISPyB picker ID was found")
    except sqlalchemy.exc.NoResultFound:
        proc_params = None

    for m in movies:
        parts = Path(m.path).parts
        visit_idx = parts.index(visit_name)
        core = Path("/") / Path(*parts[: visit_idx + 1])
        ppath = Path("/") / Path(*parts)
        sub_dataset = ppath.relative_to(core).parts[0]
        extra_path = machine_config.processed_extra_directory
        for i, p in enumerate(ppath.parts):
            if p.startswith("raw"):
                movies_path_index = i
                break
        else:
            raise ValueError(f"{m.path} does not contain a raw directory")
        mrc_out = (
            core
            / machine_config.processed_directory_name
            / sub_dataset
            / extra_path
            / "MotionCorr"
            / "job002"
            / "Movies"
            / "/".join(ppath.parts[movies_path_index + 1 : -1])
            / str(ppath.stem + "_motion_corrected.mrc")
        )
        if proc_params:

            detached_ids = [c.id for c in collected_ids]

            if not mrc_out.parent.exists():
                mrc_out.parent.mkdir(parents=True, exist_ok=True)
            if not Path(m.path).is_file():
                continue
            zocalo_message = {
                "recipes": ["em-spa-preprocess"],
                "parameters": {
                    "feedback_queue": _transport_object.feedback_queue,
                    "node_creator_queue": machine_config.node_creator_queue,
                    "dcid": detached_ids[1],
                    "kv": proc_params["voltage"],
                    "autoproc_program_id": detached_ids[3],
                    "movie": m.path,
                    "mrc_out": str(mrc_out),
                    "pixel_size": proc_params["angpix"],
                    "image_number": m.image_number,
                    "microscope": get_microscope(),
                    "mc_uuid": m.murfey_id,
                    "ft_bin": proc_params["motion_corr_binning"],
                    "fm_dose": proc_params["dose_per_frame"],
                    "gain_ref": proc_params["gain_ref"],
                    "picker_uuid": feedback_params.picker_murfey_id,
                    "session_id": args.session_id,
                    "particle_diameter": proc_params["particle_diameter"] or 0,
                    "fm_int_file": args.eer_fractionation_file,
                    "do_icebreaker_jobs": default_spa_parameters.do_icebreaker_jobs,
                },
            }
            _transport_object.send("processing_recipe", zocalo_message)
            print(f"Requested processing for {m.path}")
            time.sleep(args.injection_delay)



from __future__ import annotations

import json
import logging
import traceback
from ast import literal_eval
from pathlib import Path
from typing import Optional

from pydantic import BaseModel, field_validator
from sqlmodel import Session

from murfey.util.db import CLEMImageSeries
from murfey.workflows.clem import get_db_entry

logger = logging.getLogger("murfey.workflows.clem.register_align_and_merge_results")


class AlignAndMergeResult(BaseModel):
    series_name: str
    image_stacks: list[Path]
    align_self: Optional[str] = None
    flatten: Optional[str] = "mean"
    align_across: Optional[str] = None
    composite_image: Path

    @field_validator("image_stacks", mode="before")
    @classmethod
    def parse_stringified_list(cls, value):
        if isinstance(value, str):
            try:
                eval_result = literal_eval(value)
                if isinstance(eval_result, list):
                    parent_tiffs = [Path(p) for p in eval_result]
                    return parent_tiffs
            except (SyntaxError, ValueError):
                raise ValueError("Unable to parse input")
        # Return value as-is; if it fails, it fails
        return value


def register_align_and_merge_result(
    message: dict, murfey_db: Session, demo: bool = False
) -> bool:
    """
    session_id (recipe)
    register (wrapper)
    result (wrapper)
        key1
        key2
        ...
    """

    session_id: int = (
        int(message["session_id"])
        if not isinstance(message["session_id"], int)
        else message["session_id"]
    )

    # Validate message and try and load results
    try:
        if isinstance(message["result"], str):
            json_obj: dict = json.loads(message["result"])
            result = AlignAndMergeResult(**json_obj)
        elif isinstance(message["result"], dict):
            result = AlignAndMergeResult(**message["result"])
        else:
            logger.error(
                "Invalid type for align-and-merge processing result: "
                f"{type(message['result'])}"
            )
            return False
    except Exception:
        logger.error(
            "Exception encountered when parsing align-and-merge processing result: \n"
            f"{traceback.format_exc()}"
        )
        return False

    # Outer try-finally block for tidying up database-related section of function
    try:
        # Register items in database if not already present
        try:
            clem_img_series: CLEMImageSeries = get_db_entry(
                db=murfey_db,
                table=CLEMImageSeries,
                session_id=session_id,
                series_name=result.series_name,
            )
            clem_img_series.composite_image = str(result.composite_image)
            clem_img_series.composite_created = True
            murfey_db.add(clem_img_series)
            murfey_db.commit()

            logger.info(
                "Align-and-merge processing result registered for "
                f"{result.series_name!r} series"
            )

        except Exception:
            logger.error(
                "Exception encountered when registering align-and-merge result for "
                f"{result.series_name!r}: \n"
                f"{traceback.format_exc()}"
            )
            return False

        return True
    finally:
        murfey_db.close()



from ispyb.sqlalchemy import BLSession, Proposal
from pytest import mark
from sqlalchemy import select
from sqlalchemy.orm import Session

from murfey.server.ispyb import get_proposal_id, get_session_id
from tests.conftest import ExampleVisit


def test_get_session_id(
    ispyb_db_session: Session,
):
    # Manually get the BLSession ID for comparison
    bl_session_id = (
        ispyb_db_session.execute(
            select(BLSession)
            .join(Proposal)
            .where(BLSession.proposalId == Proposal.proposalId)
            .where(BLSession.beamLineName == ExampleVisit.instrument_name)
            .where(Proposal.proposalCode == ExampleVisit.proposal_code)
            .where(Proposal.proposalNumber == str(ExampleVisit.proposal_number))
            .where(BLSession.visit_number == ExampleVisit.visit_number)
        )
        .scalar_one()
        .sessionId
    )

    # Test function
    result = get_session_id(
        microscope=ExampleVisit.instrument_name,
        proposal_code=ExampleVisit.proposal_code,
        proposal_number=str(ExampleVisit.proposal_number),
        visit_number=str(ExampleVisit.visit_number),
        db=ispyb_db_session,
    )
    assert bl_session_id == result


def test_get_proposal_id(
    ispyb_db_session: Session,
):
    # Manually query the Proposal ID
    proposal_id = (
        ispyb_db_session.execute(
            select(Proposal)
            .where(Proposal.proposalCode == ExampleVisit.proposal_code)
            .where(Proposal.proposalNumber == str(ExampleVisit.proposal_number))
        )
        .scalar_one()
        .proposalId
    )

    # Test function
    result = get_proposal_id(
        proposal_code=ExampleVisit.proposal_code,
        proposal_number=str(ExampleVisit.proposal_number),
        db=ispyb_db_session,
    )
    assert proposal_id == result


@mark.skip
def test_get_sub_samples_from_visit():
    pass


@mark.skip
def test_get_all_ongoing_visits():
    pass



"""
CLI to generate a manifest of the FastAPI router paths present in both the instrument
server and backend server to enable lookup of the URLs based on function name.
"""

import importlib
import inspect
import pkgutil
from argparse import ArgumentParser
from pathlib import Path
from types import ModuleType
from typing import Any

import yaml
from fastapi import APIRouter

import murfey


def find_routers(name: str) -> dict[str, APIRouter]:

    def _extract_routers_from_module(module: ModuleType):
        routers = {}
        for name, obj in inspect.getmembers(module):
            if isinstance(obj, APIRouter):
                module_path = module.__name__
                key = f"{module_path}.{name}"
                routers[key] = obj
        return routers

    routers = {}

    # Import the module or package
    try:
        root = importlib.import_module(name)
    except ImportError:
        raise ImportError(
            f"Cannot import '{name}'. Please ensure that you've installed all the "
            "dependencies for the client, instrument server, and backend server "
            "before running this command."
        )

    # If it's a package, walk through submodules and extract routers from each
    if hasattr(root, "__path__"):
        module_list = pkgutil.walk_packages(root.__path__, prefix=name + ".")
        for _, module_name, _ in module_list:
            try:
                module = importlib.import_module(module_name)
            except ImportError:
                raise ImportError(
                    f"Cannot import '{module_name}'. Please ensure that you've "
                    "installed all the dependencies for the client, instrument "
                    "server, and backend server before running this command."
                )

            routers.update(_extract_routers_from_module(module))

    # Extract directly from single module
    else:
        routers.update(_extract_routers_from_module(root))

    return routers


def get_route_manifest(routers: dict[str, APIRouter]):

    manifest = {}

    for router_name, router in routers.items():
        routes = []
        for route in router.routes:
            path_params = []
            for param in route.dependant.path_params:
                param_type = param.type_ if param.type_ is not None else Any
                param_info = {
                    "name": param.name if hasattr(param, "name") else "",
                    "type": (
                        param_type.__name__
                        if hasattr(param_type, "__name__")
                        else str(param_type)
                    ),
                }
                path_params.append(param_info)
            route_info = {
                "path": route.path if hasattr(route, "path") else "",
                "function": route.name if hasattr(route, "name") else "",
                "path_params": path_params,
                "methods": list(route.methods) if hasattr(route, "methods") else [],
            }
            routes.append(route_info)
        manifest[router_name] = routes
    return manifest


def run():
    # Set up additional args
    parser = ArgumentParser()
    parser.add_argument(
        "--debug",
        action="store_true",
        default=False,
        help=("Outputs the modules being inspected when creating the route manifest"),
    )
    args = parser.parse_args()

    # Find routers
    print("Finding routers...")
    routers = {
        **find_routers("murfey.instrument_server.api"),
        **find_routers("murfey.server.api"),
    }
    # Generate the manifest
    print("Extracting route information")
    manifest = get_route_manifest(routers)

    # Verify
    if args.debug:
        for router_name, routes in manifest.items():
            print(f"Routes found in {router_name!r}")
            for route in routes:
                for key, value in route.items():
                    print(f"\t{key}: {value}")
                print()

    # Save the manifest
    murfey_dir = Path(murfey.__path__[0])
    manifest_file = murfey_dir / "util" / "route_manifest.yaml"
    with open(manifest_file, "w") as file:
        yaml.dump(manifest, file, default_flow_style=False, sort_keys=False)
    print(
        "Route manifest for instrument and backend servers saved to "
        f"{str(manifest_file)!r}"
    )
    exit()


if __name__ == "__main__":
    run()



from __future__ import annotations

import secrets
import subprocess
import time
from datetime import datetime
from functools import partial
from logging import getLogger
from pathlib import Path
from threading import Lock
from typing import Annotated, Any, Optional
from urllib.parse import urlparse

import requests
from fastapi import APIRouter, Depends, HTTPException, status
from fastapi.security import OAuth2PasswordBearer
from jose import JWTError, jwt
from pydantic import BaseModel
from werkzeug.utils import secure_filename

from murfey.client.multigrid_control import MultigridController
from murfey.client.rsync import RSyncer
from murfey.client.watchdir_multigrid import MultigridDirWatcher
from murfey.util import posix_path, sanitise, sanitise_nonpath, secure_path
from murfey.util.api import url_path_for
from murfey.util.client import read_config
from murfey.util.instrument_models import MultigridWatcherSpec
from murfey.util.models import File, Token

logger = getLogger("murfey.instrument_server.api")

watchers: dict[str | int, MultigridDirWatcher] = {}
rsyncers: dict[str, RSyncer] = {}
controllers: dict[int, MultigridController] = {}
controller_lock = Lock()
data_collection_parameters: dict = {}
tokens = {}

config = read_config()

SECRET_KEY = config["Murfey"].get("auth_key", secrets.token_hex(32))
launch_time = time.time()

encoded_jwt = jwt.encode(
    {"timestamp": launch_time},
    SECRET_KEY,
    algorithm=config["Murfey"].get("auth_algorithm", "HS256"),
)

oauth2_scheme = OAuth2PasswordBearer(tokenUrl="token")


def validate_session_token(
    session_id: int, token: Annotated[str, Depends(oauth2_scheme)]
):
    """
    Validates the token received from the backend server
    """
    try:
        decoded_data = jwt.decode(
            token,
            SECRET_KEY,
            algorithms=[config["Murfey"].get("auth_algorithm", "HS256")],
        )
        if not decoded_data.get("session") == session_id:
            raise JWTError
    except JWTError:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Could not validate credentials from backend",
            headers={"WWW-Authenticate": "Bearer"},
        )
    return session_id


MurfeySessionID = Annotated[int, Depends(validate_session_token)]

router = APIRouter()


@router.get("/health")
def health():
    return True


def _get_murfey_url() -> str:
    known_server = config["Murfey"].get("server")
    if not known_server:
        exit("Murfey server not set")
    if not known_server.startswith(("http://", "https://")):
        if "://" in known_server:
            exit("Unknown server protocol. Only http:// and https:// are allowed")
        known_server = f"http://{known_server}"
    return known_server


async def murfey_server_handshake(token: str, session_id: int | None = None) -> bool:
    # test provided token against Murfey server
    murfey_url = urlparse(_get_murfey_url(), allow_fragments=False)
    handshake_response = requests.get(
        f"{murfey_url.geturl()}{url_path_for('auth.router', 'simple_token_validation')}",
        headers={"Authorization": f"Bearer {token}"},
    )
    res = handshake_response.status_code == 200 and handshake_response.json().get(
        "valid"
    )
    if res:
        tokens["token" if session_id is None else session_id] = token
    return res


@router.post("/token")
async def token_handshake(token: Token):
    handshake_success = await murfey_server_handshake(token.access_token)
    if handshake_success:
        return Token(access_token=encoded_jwt, token_type="bearer")
    raise HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail="Handshake failure between Murfey servers",
    )


@router.post("/sessions/{session_id}/token")
async def token_handshake_for_session(session_id: int, token: Token):
    handshake_success = await murfey_server_handshake(
        token.access_token, session_id=session_id
    )
    if handshake_success:
        session_jwt = jwt.encode(
            {"session": session_id},
            SECRET_KEY,
            algorithm=config["Murfey"].get("auth_algorithm", "HS256"),
        )
        return Token(access_token=session_jwt, token_type="bearer")
    raise HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail="Handshake failure between Murfey servers",
    )


@router.get("/sessions/{session_id}/check_token")
def check_token(session_id: MurfeySessionID):
    return {"token_valid": True}


@router.post("/sessions/{session_id}/multigrid_watcher")
def setup_multigrid_watcher(
    session_id: MurfeySessionID, watcher_spec: MultigridWatcherSpec
):
    # Remove dormant controllers from memory
    with controller_lock:
        controllers_to_remove = [
            sid for sid, controller in controllers.items() if controller.dormant
        ]
        for sid in controllers_to_remove:
            del controllers[sid]

    # Return 'True' if controllers are already set up
    if controllers.get(session_id) is not None:
        return {"success": True}

    # Load machine config as dictionary
    machine_config: dict[str, Any] = requests.get(
        f"{_get_murfey_url()}{url_path_for('session_control.router', 'machine_info_by_instrument', instrument_name=sanitise_nonpath(watcher_spec.instrument_name))}",
        headers={"Authorization": f"Bearer {tokens[session_id]}"},
    ).json()

    # Set up the multigrid controller
    label = watcher_spec.label
    controllers[session_id] = MultigridController(
        [],
        watcher_spec.visit,
        watcher_spec.instrument_name,
        session_id,
        murfey_url=_get_murfey_url(),
        demo=True,
        do_transfer=True,
        processing_enabled=not watcher_spec.skip_existing_processing,
        _machine_config=machine_config,
        token=tokens.get(session_id, "token"),
        data_collection_parameters=data_collection_parameters.get(label, {}),
        rsync_restarts=watcher_spec.rsync_restarts,
        visit_end_time=watcher_spec.visit_end_time,
    )
    # Make child directories, if specified
    watcher_spec.source.mkdir(exist_ok=True)
    for d in machine_config.get("create_directories", []):
        (watcher_spec.source / d).mkdir(exist_ok=True)

    # Set up multigrid directory watcher
    watchers[session_id] = MultigridDirWatcher(
        watcher_spec.source,
        machine_config,
        skip_existing_processing=watcher_spec.skip_existing_processing,
    )
    watchers[session_id].subscribe(
        partial(
            controllers[session_id]._start_rsyncer_multigrid,
            destination_overrides=watcher_spec.destination_overrides,
        )
    )
    watchers[session_id].subscribe(
        controllers[session_id]._multigrid_watcher_finalised, final=True
    )
    return {"success": True}


@router.post("/sessions/{session_id}/start_multigrid_watcher")
def start_multigrid_watcher(session_id: MurfeySessionID, process: bool = True):
    if watchers.get(session_id) is None:
        return {"success": False}
    if not process:
        watchers[session_id]._analyse = False
    try:
        watchers[session_id].start()
    # Ignore RuntimeError; this happens when reconnecting after a backend server restart
    except RuntimeError:
        logger.debug(f"MultigridWatcher for session {session_id} is already active")
    return {"success": True}


@router.delete("/sessions/{session_id}/multigrid_watcher/{label}")
def stop_multigrid_watcher(session_id: MurfeySessionID, label: str):
    watchers[label].request_stop()
    return {"success": True}


@router.get("/sessions/{session_id}/multigrid_controller/status")
def check_multigrid_controller_status(
    session_id: MurfeySessionID,
):
    if controllers.get(session_id, None) is not None:
        return {
            "dormant": controllers[session_id].dormant,
            "exists": True,
            "finalising": controllers[session_id].finalising,
        }
    return {"exists": False}


@router.post("/sessions/{session_id}/multigrid_controller/visit_end_time")
def update_multigrid_controller_visit_end_time(
    session_id: MurfeySessionID, end_time: datetime
):
    controllers[session_id].update_visit_time(end_time)
    return {"success": True}


class RsyncerSource(BaseModel):
    source: Path


@router.post("/sessions/{session_id}/stop_rsyncer")
def stop_rsyncer(session_id: MurfeySessionID, rsyncer_source: RsyncerSource):
    controllers[session_id].rsync_processes[rsyncer_source.source]._halt_thread = True
    return {"success": True}


@router.post("/sessions/{session_id}/remove_rsyncer")
def remove_rsyncer(session_id: MurfeySessionID, rsyncer_source: RsyncerSource):
    controllers[session_id]._request_watcher_stop(rsyncer_source.source)
    controllers[session_id].rsync_processes[rsyncer_source.source]._stopping = True
    controllers[session_id].rsync_processes[rsyncer_source.source]._halt_thread = True
    controllers[session_id].rsync_processes[rsyncer_source.source].queue.put(
        None, block=False
    )
    return {"success": True}


@router.post("/sessions/{session_id}/abandon_controller")
def abandon_controller(session_id: MurfeySessionID):
    controllers[session_id].abandon()
    return {"success": True}


@router.post("/sessions/{session_id}/finalise_rsyncer")
def finalise_rsyncer(session_id: MurfeySessionID, rsyncer_source: RsyncerSource):
    controllers[session_id]._finalise_rsyncer(rsyncer_source.source)
    return {"success": True}


@router.post("/sessions/{session_id}/finalise_session")
def finalise_session(session_id: MurfeySessionID):
    watchers[session_id].request_stop()
    logger.debug(f"Stop request sent to multigrid watcher for session {session_id}")
    controllers[session_id].finalise()
    logger.debug(f"Stop orders sent to multigrid controller for session {session_id} ")
    return {"success": True}


@router.post("/sessions/{session_id}/restart_rsyncer")
def restart_rsyncer(session_id: MurfeySessionID, rsyncer_source: RsyncerSource):
    controllers[session_id]._restart_rsyncer(rsyncer_source.source)
    return {"success": True}


@router.post("/sessions/{session_id}/flush_skipped_rsyncer")
def flush_skipped_rsyncer(session_id: MurfeySessionID, rsyncer_source: RsyncerSource):
    controllers[session_id].rsync_processes[rsyncer_source.source].flush_skipped()
    return {"success": True}


class ObserverInfo(BaseModel):
    source: str
    num_files_transferred: int
    num_files_in_queue: int
    alive: bool
    stopping: bool
    num_files_skipped: int = 0


@router.get("/sessions/{session_id}/rsyncer_info")
def get_rsyncer_info(session_id: MurfeySessionID) -> list[ObserverInfo]:
    info: list[ObserverInfo] = []
    if controllers.get(session_id, None) is None:
        logger.debug(f"Multigrid controller for session {session_id} doesn't exist")
        return info
    for k, v in controllers[session_id].rsync_processes.items():
        info.append(
            ObserverInfo(
                source=str(k),
                num_files_transferred=v._files_transferred,
                num_files_in_queue=v.queue.qsize(),
                alive=v.thread.is_alive(),
                stopping=v._stopping,
                num_files_skipped=len(v._skipped_files),
            )
        )
    return info


@router.get("/sessions/{session_id}/analyser_info")
def get_analyser_info(session_id: MurfeySessionID) -> list[ObserverInfo]:
    info: list[ObserverInfo] = []
    if controllers.get(session_id, None) is None:
        logger.debug(f"Multigrid controller for session {session_id} doesn't exist")
        return info
    for k, v in controllers[session_id].analysers.items():
        info.append(
            ObserverInfo(
                source=str(k),
                num_files_transferred=0,
                num_files_in_queue=v.queue.qsize(),
                alive=v.thread.is_alive(),
                stopping=v._stopping,
            )
        )
    return info


class ProcessingParameters(BaseModel):
    gain_ref: str
    dose_per_frame: Optional[float] = None
    extract_downscale: bool = True
    particle_diameter: Optional[float] = None
    symmetry: str = "C1"
    eer_fractionation: int = 20


class ProcessingParameterBlock(BaseModel):
    label: str
    params: ProcessingParameters


@router.post("/sessions/{session_id}/processing_parameters")
def register_processing_parameters(
    session_id: MurfeySessionID, proc_param_block: ProcessingParameterBlock
):
    data_collection_parameters[proc_param_block.label] = {}
    for k, v in proc_param_block.params.model_dump().items():
        if v is not None:
            data_collection_parameters[proc_param_block.label][k] = v
    if controllers.get(session_id):
        controllers[session_id].data_collection_parameters.update(
            data_collection_parameters[proc_param_block.label]
        )
        for k, v in proc_param_block.params.model_dump().items():
            if v is not None and hasattr(controllers[session_id]._environment, k):
                setattr(controllers[session_id]._environment, k, v)
    return {"success": True}


@router.get(
    "/instruments/{instrument_name}/sessions/{session_id}/possible_gain_references"
)
def get_possible_gain_references(
    instrument_name: str, session_id: MurfeySessionID
) -> list[File]:
    machine_config = requests.get(
        f"{_get_murfey_url()}{url_path_for('session_control.router', 'machine_info_by_instrument', instrument_name=sanitise_nonpath(instrument_name))}",
        headers={"Authorization": f"Bearer {tokens[session_id]}"},
    ).json()
    candidates = []
    for gf in secure_path(
        Path(machine_config["gain_reference_directory"]), keep_spaces=True
    ).glob("**/*"):
        if gf.is_file():
            candidates.append(
                File(
                    name=gf.name,
                    description="",
                    size=gf.stat().st_size / 1e6,
                    timestamp=datetime.fromtimestamp(gf.stat().st_mtime),
                    full_path=str(gf),
                )
            )
    candidates.sort(key=lambda x: x.timestamp, reverse=True)
    return candidates


class GainReference(BaseModel):
    gain_path: Path
    visit_path: str
    gain_destination_dir: str = "processing"


@router.post(
    "/instruments/{instrument_name}/sessions/{session_id}/upload_gain_reference"
)
def upload_gain_reference(
    instrument_name: str, session_id: MurfeySessionID, gain_reference: GainReference
):
    safe_gain_path = sanitise(str(gain_reference.gain_path))
    safe_visit_path = sanitise(gain_reference.visit_path)
    safe_destination_dir = sanitise(gain_reference.gain_destination_dir)

    # Load machine config and other needed properties
    machine_config: dict[str, Any] = requests.get(
        f"{_get_murfey_url()}{url_path_for('session_control.router', 'machine_info_by_instrument', instrument_name=sanitise_nonpath(instrument_name))}",
        headers={"Authorization": f"Bearer {tokens[session_id]}"},
    ).json()

    # Validate that file passed is from the gain reference directory
    gain_ref_dir = machine_config.get("gain_reference_directory", "")
    if not safe_gain_path.startswith(gain_ref_dir):
        raise ValueError(
            "Gain reference file does not originate from the gain reference directory "
            f"{gain_ref_dir!r}"
        )

    # Return the rsync URL if set, otherwise assume you are syncing via Murfey
    rsync_url = urlparse(
        str(machine_config["rsync_url"])
        if machine_config.get("rsync_url", "")
        else _get_murfey_url()
    )
    rsync_module = machine_config.get("rsync_module", "data")
    rsync_path = f"{rsync_url.hostname}::{rsync_module}/{safe_visit_path}/{safe_destination_dir}/{secure_filename(gain_reference.gain_path.name)}"

    # Run rsync subprocess to transfer gain reference
    cmd = [
        "rsync",
        posix_path(Path(safe_gain_path)),
        rsync_path,
    ]
    gain_rsync = subprocess.run(
        cmd,
        capture_output=True,
        text=True,
    )
    if gain_rsync.returncode:
        logger.warning(
            f"Failed to transfer gain reference file {safe_gain_path!r} to {f'{safe_visit_path}/processing'!r} \n"
            f"Executed the following command: {' '.join(cmd)!r} \n"
            f"Returned the following error: \n"
            f"{gain_rsync.stderr}"
        )
        return {"success": False}
    return {"success": True}


class UpstreamTiffInfo(BaseModel):
    download_dir: Path


@router.post("/visits/{visit_name}/sessions/{session_id}/upstream_tiff_data_request")
def gather_upstream_tiffs(
    visit_name: str, session_id: MurfeySessionID, upstream_tiff_info: UpstreamTiffInfo
):
    sanitised_visit_name = sanitise_nonpath(visit_name)
    assert not any(c in visit_name for c in ("/", "\\", ":", ";"))
    murfey_url = urlparse(_get_murfey_url(), allow_fragments=False)
    upstream_tiff_info.download_dir.mkdir(exist_ok=True)
    upstream_tiff_paths = (
        requests.get(
            f"{murfey_url.geturl()}{url_path_for('session_control.correlative_router', 'gather_upstream_tiffs', session_id=session_id, visit_name=sanitised_visit_name)}",
            headers={"Authorization": f"Bearer {tokens[session_id]}"},
        ).json()
        or []
    )
    for tiff_path in upstream_tiff_paths:
        tiff_data = requests.get(
            f"{murfey_url.geturl()}{url_path_for('session_control.correlative_router', 'get_tiff', session_id=session_id, visit_name=sanitised_visit_name, tiff_path=tiff_path)}",
            stream=True,
            headers={"Authorization": f"Bearer {tokens[session_id]}"},
        )
        with open(upstream_tiff_info.download_dir / tiff_path, "wb") as utiff:
            for chunk in tiff_data.iter_content(chunk_size=32 * 1024**2):
                utiff.write(chunk)



import logging

import numpy as np
from sqlmodel import Session, select

from murfey.server import _transport_object
from murfey.server.api.auth import MurfeySessionIDInstrument as MurfeySessionID
from murfey.server.gain import Camera
from murfey.util import sanitise
from murfey.util.config import get_machine_config
from murfey.util.db import DataCollectionGroup, SearchMap
from murfey.util.db import Session as MurfeySession
from murfey.util.db import TiltSeries
from murfey.util.models import BatchPositionParameters, SearchMapParameters

logger = logging.getLogger("murfey.client.util.tomo_metadata")


def register_search_map_in_database(
    session_id: MurfeySessionID,
    search_map_name: str,
    search_map_params: SearchMapParameters,
    murfey_db: Session,
    close_db: bool = True,
):
    dcg = murfey_db.exec(
        select(DataCollectionGroup)
        .where(DataCollectionGroup.session_id == session_id)
        .where(DataCollectionGroup.tag == search_map_params.tag)
    ).one()
    try:
        # See if there is already a search map with this name and update if so
        search_map = murfey_db.exec(
            select(SearchMap)
            .where(SearchMap.name == search_map_name)
            .where(SearchMap.tag == search_map_params.tag)
            .where(SearchMap.session_id == session_id)
        ).one()
        search_map.x_stage_position = (
            search_map_params.x_stage_position or search_map.x_stage_position
        )
        search_map.y_stage_position = (
            search_map_params.y_stage_position or search_map.y_stage_position
        )
        search_map.pixel_size = search_map_params.pixel_size or search_map.pixel_size
        search_map.image = search_map_params.image or search_map.image
        search_map.binning = search_map_params.binning or search_map.binning
        search_map.reference_matrix_m11 = (
            search_map_params.reference_matrix.get("m11")
            or search_map.reference_matrix_m11
        )
        search_map.reference_matrix_m12 = (
            search_map_params.reference_matrix.get("m12")
            or search_map.reference_matrix_m12
        )
        search_map.reference_matrix_m21 = (
            search_map_params.reference_matrix.get("m21")
            or search_map.reference_matrix_m21
        )
        search_map.reference_matrix_m22 = (
            search_map_params.reference_matrix.get("m22")
            or search_map.reference_matrix_m22
        )
        search_map.stage_correction_m11 = (
            search_map_params.stage_correction.get("m11")
            or search_map.stage_correction_m11
        )
        search_map.stage_correction_m12 = (
            search_map_params.stage_correction.get("m12")
            or search_map.stage_correction_m12
        )
        search_map.stage_correction_m21 = (
            search_map_params.stage_correction.get("m21")
            or search_map.stage_correction_m21
        )
        search_map.stage_correction_m22 = (
            search_map_params.stage_correction.get("m22")
            or search_map.stage_correction_m22
        )
        search_map.image_shift_correction_m11 = (
            search_map_params.image_shift_correction.get("m11")
            or search_map.image_shift_correction_m11
        )
        search_map.image_shift_correction_m12 = (
            search_map_params.image_shift_correction.get("m12")
            or search_map.image_shift_correction_m12
        )
        search_map.image_shift_correction_m21 = (
            search_map_params.image_shift_correction.get("m21")
            or search_map.image_shift_correction_m21
        )
        search_map.image_shift_correction_m22 = (
            search_map_params.image_shift_correction.get("m22")
            or search_map.image_shift_correction_m22
        )
        search_map.height = search_map_params.height or search_map.height
        search_map.width = search_map_params.width or search_map.width
        if _transport_object:
            _transport_object.do_update_search_map(search_map.id, search_map_params)
    except Exception as e:
        logger.info(f"Registering new search map due to {e}", exc_info=True)
        if _transport_object:
            sm_ispyb_response = _transport_object.do_insert_search_map(
                dcg.atlas_id, search_map_params
            )
        else:
            # mock up response so that below still works
            sm_ispyb_response = {"success": False, "return_value": None}
        # Register new search map
        search_map = SearchMap(
            id=(
                sm_ispyb_response["return_value"]
                if sm_ispyb_response["success"]
                else None
            ),
            name=search_map_name,
            session_id=session_id,
            tag=search_map_params.tag,
            x_stage_position=search_map_params.x_stage_position,
            y_stage_position=search_map_params.y_stage_position,
            pixel_size=search_map_params.pixel_size,
            image=search_map_params.image,
            binning=search_map_params.binning,
            reference_matrix_m11=search_map_params.reference_matrix.get("m11"),
            reference_matrix_m12=search_map_params.reference_matrix.get("m12"),
            reference_matrix_m21=search_map_params.reference_matrix.get("m21"),
            reference_matrix_m22=search_map_params.reference_matrix.get("m22"),
            stage_correction_m11=search_map_params.stage_correction.get("m11"),
            stage_correction_m12=search_map_params.stage_correction.get("m12"),
            stage_correction_m21=search_map_params.stage_correction.get("m21"),
            stage_correction_m22=search_map_params.stage_correction.get("m22"),
            image_shift_correction_m11=search_map_params.image_shift_correction.get(
                "m11"
            ),
            image_shift_correction_m12=search_map_params.image_shift_correction.get(
                "m12"
            ),
            image_shift_correction_m21=search_map_params.image_shift_correction.get(
                "m21"
            ),
            image_shift_correction_m22=search_map_params.image_shift_correction.get(
                "m22"
            ),
            height=search_map_params.height,
            width=search_map_params.width,
        )

    murfey_session = murfey_db.exec(
        select(MurfeySession).where(MurfeySession.id == session_id)
    ).one()
    machine_config = get_machine_config(instrument_name=murfey_session.instrument_name)[
        murfey_session.instrument_name
    ]
    if all(
        [
            search_map.reference_matrix_m11,
            search_map.stage_correction_m11,
            search_map.x_stage_position,
            search_map.y_stage_position,
            search_map.pixel_size,
            search_map.height,
            search_map.width,
            dcg.atlas_pixel_size,
        ]
    ):
        # Work out the shifted positions if all required information is present
        reference_shift_matrix = np.array(
            [
                [
                    search_map.reference_matrix_m11,
                    search_map.reference_matrix_m12,
                ],
                [
                    search_map.reference_matrix_m21,
                    search_map.reference_matrix_m22,
                ],
            ]
        )
        stage_vector = np.array(
            [search_map.x_stage_position, search_map.y_stage_position]
        )
        stage_correction_matrix = np.array(
            [
                [
                    search_map.stage_correction_m11,
                    search_map.stage_correction_m12,
                ],
                [
                    search_map.stage_correction_m21,
                    search_map.stage_correction_m22,
                ],
            ]
        )
        corrected_vector = np.matmul(
            np.linalg.inv(reference_shift_matrix),
            np.matmul(
                stage_correction_matrix, np.matmul(reference_shift_matrix, stage_vector)
            ),
        )

        # Flip positions based on camera type
        camera = getattr(Camera, machine_config.camera)
        if camera == Camera.FALCON:
            # Flip in y
            corrected_vector = np.matmul(np.array([[1, 0], [0, -1]]), corrected_vector)
        elif camera == Camera.K3_FLIPX:
            # Flip in x
            corrected_vector = np.matmul(np.array([[-1, 0], [0, 1]]), corrected_vector)
        elif camera == Camera.K3_FLIPY:
            # Inversion (camera mounted upside-down)
            corrected_vector = np.matmul(np.array([[0, 1], [1, 0]]), corrected_vector)

        # Convert from metres to pixels
        search_map_params.height_on_atlas = int(
            search_map.height * search_map.pixel_size / dcg.atlas_pixel_size
        )
        search_map_params.width_on_atlas = int(
            search_map.width * search_map.pixel_size / dcg.atlas_pixel_size
        )
        search_map_params.x_location = float(
            corrected_vector[0] / dcg.atlas_pixel_size + 2003
        )
        search_map_params.y_location = float(
            corrected_vector[1] / dcg.atlas_pixel_size + 2003
        )
        search_map.x_location = search_map_params.x_location
        search_map.y_location = search_map_params.y_location
        if _transport_object:
            _transport_object.do_update_search_map(search_map.id, search_map_params)
    else:
        logger.info(
            f"Unable to register search map {sanitise(search_map_name)} position yet: "
            f"stage {sanitise(str(search_map_params.x_stage_position))}, "
            f"width {sanitise(str(search_map_params.width))}, "
            f"atlas pixel size {sanitise(str(dcg.atlas_pixel_size))}"
        )
    murfey_db.add(search_map)
    murfey_db.commit()
    if close_db:
        murfey_db.close()


def register_batch_position_in_database(
    session_id: MurfeySessionID,
    batch_name: str,
    batch_parameters: BatchPositionParameters,
    murfey_db: Session,
):
    search_map = murfey_db.exec(
        select(SearchMap)
        .where(SearchMap.name == batch_parameters.search_map_name)
        .where(SearchMap.tag == batch_parameters.tag)
        .where(SearchMap.session_id == session_id)
    ).one()

    try:
        tilt_series = murfey_db.exec(
            select(TiltSeries)
            .where(TiltSeries.tag == batch_name)
            .where(TiltSeries.rsync_source == batch_parameters.tag)
            .where(TiltSeries.session_id == session_id)
        ).one()
        if tilt_series.x_location:
            logger.info(
                f"Already did position analysis for tomogram {sanitise(batch_name)}"
            )
            return
    except Exception:
        tilt_series = TiltSeries(
            tag=batch_name,
            rsync_source=batch_parameters.tag,
            session_id=session_id,
            search_map_id=search_map.id,
        )

    # Get the pixel location on the searchmap
    if all(
        [
            search_map.reference_matrix_m11,
            search_map.stage_correction_m11,
            search_map.x_stage_position,
            search_map.y_stage_position,
            search_map.pixel_size,
            search_map.height,
            search_map.width,
        ]
    ):
        reference_shift_matrix = np.array(
            [
                [
                    search_map.reference_matrix_m11,
                    search_map.reference_matrix_m12,
                ],
                [
                    search_map.reference_matrix_m21,
                    search_map.reference_matrix_m22,
                ],
            ]
        )
        stage_correction_matrix = np.array(
            [
                [
                    search_map.stage_correction_m11,
                    search_map.stage_correction_m12,
                ],
                [
                    search_map.stage_correction_m21,
                    search_map.stage_correction_m22,
                ],
            ]
        )
        image_shift_matrix = np.array(
            [
                [
                    search_map.image_shift_correction_m11,
                    search_map.image_shift_correction_m12,
                ],
                [
                    search_map.image_shift_correction_m21,
                    search_map.image_shift_correction_m22,
                ],
            ]
        )

        stage_vector = np.array(
            [
                batch_parameters.x_stage_position - search_map.x_stage_position,
                batch_parameters.y_stage_position - search_map.y_stage_position,
            ]
        )

        # Apply stage corrections to the location
        corrected_vector = np.matmul(
            np.linalg.inv(reference_shift_matrix),
            np.matmul(
                np.linalg.inv(stage_correction_matrix),
                np.matmul(
                    np.linalg.inv(image_shift_matrix),
                    np.matmul(reference_shift_matrix, stage_vector),
                ),
            ),
        )
        # Invert y coordinate
        inverted_corrected_vector = np.matmul([[1, 0], [0, -1]], corrected_vector)
        # Apply shift from centre
        centre_batch_pixel = inverted_corrected_vector / search_map.pixel_size + [
            search_map.width / 2,
            search_map.height / 2,
        ]
        tilt_series.x_location = (
            (
                centre_batch_pixel[0]
                - batch_parameters.x_beamshift / search_map.pixel_size
            )
            * 512
            / search_map.width
        )
        tilt_series.y_location = (
            (
                centre_batch_pixel[1]
                + batch_parameters.y_beamshift / search_map.pixel_size
            )
            * 512
            / search_map.width
        )
    else:
        logger.warning(
            f"Incomplete search map for position of {sanitise(batch_name)}: "
            f"stage {search_map.x_stage_position}, "
            f"width {search_map.width}, "
        )
    murfey_db.add(tilt_series)
    murfey_db.commit()



from logging import getLogger
from typing import Optional

import sqlalchemy
from fastapi import APIRouter, Depends
from pydantic import BaseModel
from sqlmodel import Session, select

from murfey.server.api.auth import MurfeySessionIDFrontend as MurfeySessionID
from murfey.server.api.auth import validate_token
from murfey.server.murfey_db import murfey_db
from murfey.util.db import SessionProcessingParameters

logger = getLogger("murfey.server.api.processing_parameters")

router = APIRouter(
    prefix="/session_parameters",
    dependencies=[Depends(validate_token)],
    tags=["Processing Parameters"],
)


class EditableSessionProcessingParameters(BaseModel):
    gain_ref: str = ""
    dose_per_frame: Optional[float] = None
    eer_fractionation_file: str = ""
    symmetry: str = ""


@router.get("/sessions/{session_id}/session_processing_parameters")
def get_session_processing_parameters(
    session_id: MurfeySessionID, db: Session = murfey_db
) -> Optional[EditableSessionProcessingParameters]:
    try:
        proc_params = db.exec(
            select(SessionProcessingParameters).where(
                SessionProcessingParameters.session_id == session_id
            )
        ).one()
    except sqlalchemy.exc.NoResultFound:
        return None
    return EditableSessionProcessingParameters(
        gain_ref=proc_params.gain_ref,
        dose_per_frame=proc_params.dose_per_frame,
        eer_fractionation_file=proc_params.eer_fractionation_file,
        symmetry=proc_params.symmetry,
    )


@router.post("/sessions/{session_id}/session_processing_parameters")
def set_session_processing_parameters(
    session_id: MurfeySessionID,
    edited_parameters: EditableSessionProcessingParameters,
    db: Session = murfey_db,
) -> EditableSessionProcessingParameters:
    proc_params = db.exec(
        select(SessionProcessingParameters).where(
            SessionProcessingParameters.session_id == session_id
        )
    ).one()
    proc_params.gain_ref = edited_parameters.gain_ref or proc_params.gain_ref
    proc_params.dose_per_frame = (
        edited_parameters.dose_per_frame or proc_params.dose_per_frame
    )
    proc_params.eer_fractionation_file = (
        edited_parameters.eer_fractionation_file or proc_params.eer_fractionation_file
    )
    proc_params.symmetry = edited_parameters.symmetry or proc_params.symmetry
    db.add(proc_params)
    db.commit()
    db.close()
    return edited_parameters



from __future__ import annotations

import math
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, field_validator

"""
General Models
==============
Models used in multiple workflows.
"""


class Visit(BaseModel):
    start: datetime
    end: datetime
    session_id: int
    name: str
    beamline: str
    proposal_title: str

    def __repr__(self) -> str:
        return (
            "Visit("
            f"start='{self.start:%Y-%m-%d %H:%M}', "
            f"end='{self.end:%Y-%m-%d %H:%M}', "
            f"session_id='{self.session_id!r}',"
            f"name={self.name!r}, "
            f"beamline={self.beamline!r}, "
            f"proposal_title={self.proposal_title!r}"
            ")"
        )


class RegistrationMessage(BaseModel):
    registration: str
    params: Optional[Dict[str, Any]] = None


class File(BaseModel):
    name: str
    description: str
    size: int
    timestamp: datetime
    full_path: str

    @field_validator("size", mode="before")
    @classmethod
    def round_file_size_correctly(cls, v: Any) -> int:
        if isinstance(v, float):
            if v - math.floor(v) == 0.5:
                return math.ceil(v)
            return round(v)
        return v


class ConnectionFileParameters(BaseModel):
    filename: str
    destinations: List[str]


class ClientInfo(BaseModel):
    id: int


class RsyncerInfo(BaseModel):
    source: str
    destination: str
    session_id: int
    transferring: bool = True
    increment_count: int = 1
    bytes: int = 0
    increment_data_count: int = 0
    data_bytes: int = 0
    tag: str = ""


"""
Single Particle Analysis
========================
Models related to the single-particle analysis workflow.
"""


class ProcessingParametersSPA(BaseModel):
    tag: str
    dose_per_frame: Optional[float] = None
    gain_ref: Optional[str] = None
    experiment_type: str
    voltage: float
    image_size_x: int
    image_size_y: int
    pixel_size_on_image: str
    motion_corr_binning: int
    file_extension: str
    acquisition_software: str
    symmetry: str
    eer_fractionation_file: str = ""
    magnification: Optional[int] = None
    total_exposed_dose: Optional[float] = None
    c2aperture: Optional[float] = None
    exposure_time: Optional[float] = None
    slit_width: Optional[float] = None
    phase_plate: bool = False

    class Base(BaseModel):
        dose_per_frame: Optional[float] = None
        gain_ref: Optional[str] = None
        symmetry: str
        eer_fractionation: int


class GridSquareParameters(BaseModel):
    tag: str
    x_location: Optional[float] = None
    y_location: Optional[float] = None
    x_stage_position: Optional[float] = None
    y_stage_position: Optional[float] = None
    readout_area_x: Optional[int] = None
    readout_area_y: Optional[int] = None
    thumbnail_size_x: Optional[int] = None
    thumbnail_size_y: Optional[int] = None
    height: Optional[int] = None
    width: Optional[int] = None
    pixel_size: Optional[float] = None
    image: str = ""
    angle: Optional[float] = None


class FoilHoleParameters(BaseModel):
    tag: str
    name: int
    x_location: Optional[float] = None
    y_location: Optional[float] = None
    x_stage_position: Optional[float] = None
    y_stage_position: Optional[float] = None
    readout_area_x: Optional[int] = None
    readout_area_y: Optional[int] = None
    thumbnail_size_x: Optional[int] = None
    thumbnail_size_y: Optional[int] = None
    pixel_size: Optional[float] = None
    image: str = ""
    diameter: Optional[float] = None


class SearchMapParameters(BaseModel):
    tag: str
    x_location: Optional[float] = None
    y_location: Optional[float] = None
    x_stage_position: Optional[float] = None
    y_stage_position: Optional[float] = None
    pixel_size: Optional[float] = None
    image: Optional[str] = None
    binning: Optional[float] = None
    reference_matrix: Dict[str, float] = {}
    stage_correction: Dict[str, float] = {}
    image_shift_correction: Dict[str, float] = {}
    height: Optional[int] = None
    width: Optional[int] = None
    height_on_atlas: Optional[int] = None
    width_on_atlas: Optional[int] = None


class BatchPositionParameters(BaseModel):
    tag: str
    x_stage_position: float
    y_stage_position: float
    x_beamshift: float
    y_beamshift: float
    search_map_name: str


class MultigridWatcherSetup(BaseModel):
    source: Path
    skip_existing_processing: bool = False
    destination_overrides: Dict[Path, str] = {}
    rsync_restarts: List[str] = []


class Token(BaseModel):
    access_token: str
    token_type: str


"""
Tomography
==========
Models related to the tomographic reconstruction workflow.
"""


class ProcessingParametersTomo(BaseModel):
    dose_per_frame: Optional[float] = None
    frame_count: int
    tilt_axis: float
    gain_ref: Optional[str] = None
    experiment_type: str
    voltage: float
    image_size_x: int
    image_size_y: int
    pixel_size_on_image: str
    motion_corr_binning: int
    file_extension: str
    tag: str
    tilt_series_tag: str
    eer_fractionation_file: Optional[str] = None
    eer_fractionation: int

    class Base(BaseModel):
        dose_per_frame: Optional[float] = None
        gain_ref: Optional[str] = None
        eer_fractionation: int


class CompletedTiltSeries(BaseModel):
    tilt_series: List[str]
    rsync_source: str



import argparse

from sqlmodel import Session, create_engine, select

from murfey.server.murfey_db import url
from murfey.util.config import get_machine_config
from murfey.util.db import Session as MurfeySession


def run():
    parser = argparse.ArgumentParser(
        description="See and remove Murfey client sessions from Murfey DB"
    )

    parser.add_argument("-d", "--delete", dest="sessions_to_remove", nargs="+")

    args = parser.parse_args()

    _url = url(get_machine_config())
    engine = create_engine(_url)
    murfey_db = Session(engine)
    if args.sessions_to_remove:
        for sess_id in args.sessions_to_remove:
            sess = murfey_db.exec(
                select(MurfeySession).where(MurfeySession.id == sess_id)
            ).one()
            murfey_db.delete(sess)
        murfey_db.commit()
    else:
        sessions = murfey_db.exec(select(MurfeySession)).all()
        print(sessions)
    murfey_db.close()



from __future__ import annotations

import logging
import re
from pathlib import Path
from typing import Optional, Type, Union

from sqlalchemy.exc import NoResultFound
from sqlmodel import Session, select

from murfey.util.config import get_machine_config
from murfey.util.db import (
    CLEMImageMetadata,
    CLEMImageSeries,
    CLEMImageStack,
    CLEMLIFFile,
    CLEMTIFFFile,
)
from murfey.util.db import Session as MurfeySession

logger = logging.getLogger("murfey.workflows.clem")


"""
HELPER FUNCTIONS FOR CLEM DATABASE
"""


def _validate_and_sanitise(
    file: Path,
    session_id: int,
    db: Session,
) -> Path:
    """
    Performs validation and sanitisation on the incoming file paths, ensuring that
    no forbidden characters are present and that the the path points only to allowed
    sections of the file server.

    Returns the file path as a sanitised string that can be converted into a Path
    object again.

    NOTE: Due to the instrument name query, 'db' now needs to be passed as an
    explicit variable to this function from within a FastAPI endpoint, as using the
    instance that was imported directly won't load it in the correct state.
    """

    valid_file_types = (
        ".lif",
        ".tif",
        ".tiff",
        ".xlif",
        ".xml",
    )

    # Resolve symlinks and directory changes to get full file path
    full_path = Path(file).resolve()

    # Use machine configuration to validate which file base paths are accepted from
    instrument_name = (
        db.exec(select(MurfeySession).where(MurfeySession.id == session_id))
        .one()
        .instrument_name
    )
    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]
    rsync_basepath = machine_config.rsync_basepath.resolve()

    # Check that full file path doesn't contain unallowed characters
    # Currently allows only:
    # - words (alphanumerics and "_"; \w),
    # - spaces (\s),
    # - periods,
    # - dashes,
    # - forward slashes ("/")
    if bool(re.fullmatch(r"^[\w\s\.\-/]+$", str(full_path))) is False:
        raise ValueError(f"Unallowed characters present in {file}")

    # Check that it's not accessing somehwere it's not allowed
    if not str(full_path).startswith(str(rsync_basepath)):
        raise ValueError(f"{file} points to a directory that is not permitted")

    # Check that it is of a permitted file type
    if f"{full_path.suffix}" not in valid_file_types:
        raise ValueError(f"{full_path.suffix} is not a permitted file format")

    return full_path


def get_db_entry(
    db: Session,
    # With the database search funcion having been moved out of the FastAPI
    # endpoint, the database now has to be explicitly passed within the FastAPI
    # endpoint function in order for it to be loaded in the correct state.
    table: Type[
        Union[
            CLEMImageMetadata,
            CLEMImageSeries,
            CLEMImageStack,
            CLEMLIFFile,
            CLEMTIFFFile,
        ]
    ],
    session_id: int,
    file_path: Optional[Path] = None,
    series_name: Optional[str] = None,
) -> Union[
    CLEMImageMetadata,
    CLEMImageSeries,
    CLEMImageStack,
    CLEMLIFFile,
    CLEMTIFFFile,
]:
    """
    Searches the CLEM workflow-related tables in the Murfey database for an entry that
    matches the file path or series name within a given session. Returns the entry if
    a match is found, otherwise register it as a new entry in the database.
    """

    # Validate that parameters are provided correctly
    if file_path is None and series_name is None:
        raise ValueError(
            "One of either 'file_path' or 'series_name' has to be provided"
        )
    if file_path is not None and series_name is not None:
        raise ValueError("Only one of 'file_path' or 'series_name' should be provided")

    # Validate file path if provided
    if file_path is not None:
        try:
            file_path = _validate_and_sanitise(file_path, session_id, db)
        except Exception:
            raise Exception

    # Validate series name to use
    if series_name is not None:
        if bool(re.fullmatch(r"^[\w\s\.\-/]+$", series_name)) is False:
            raise ValueError("One or more characters in the string are not permitted")

    # Return database entry if it exists
    try:
        db_entry = (
            db.exec(
                select(table)
                .where(table.session_id == session_id)
                .where(table.file_path == str(file_path))
            ).one()
            if file_path is not None
            else db.exec(
                select(table)
                .where(table.session_id == session_id)
                .where(table.series_name == series_name)
            ).one()
        )
    # Create and register new entry if not present
    except NoResultFound:
        db_entry = (
            table(
                file_path=str(file_path),
                session_id=session_id,
            )
            if file_path is not None
            else table(
                series_name=series_name,
                session_id=session_id,
            )
        )
        db.add(db_entry)
        db.commit()

    except Exception:
        raise Exception

    return db_entry



from __future__ import annotations

from logging import getLogger
from typing import Optional

from fastapi import APIRouter, Depends
from prometheus_client import Counter, Gauge
from sqlmodel import select

import murfey.server.prometheus as prom
from murfey.server.api.auth import validate_instrument_token
from murfey.server.murfey_db import murfey_db
from murfey.util import sanitise
from murfey.util.db import RsyncInstance
from murfey.util.models import RsyncerInfo

logger = getLogger("murfey.server.api.prometheus")

router = APIRouter(
    prefix="/prometheus",
    dependencies=[Depends(validate_instrument_token)],
    tags=["Prometheus"],
)


@router.post("/visits/{visit_name}/increment_rsync_file_count")
def increment_rsync_file_count(
    visit_name: str, rsyncer_info: RsyncerInfo, db=murfey_db
):
    try:
        rsync_instance = db.exec(
            select(RsyncInstance).where(
                RsyncInstance.source == rsyncer_info.source,
                RsyncInstance.destination == rsyncer_info.destination,
                RsyncInstance.session_id == rsyncer_info.session_id,
            )
        ).one()
    except Exception:
        logger.error(
            f"Failed to find rsync instance for visit {sanitise(visit_name)} "
            "with the following properties: \n"
            f"{rsyncer_info.model_dump()}",
            exc_info=True,
        )
        return None
    rsync_instance.files_counted += rsyncer_info.increment_count
    db.add(rsync_instance)
    db.commit()
    db.close()
    prom.seen_files.labels(rsync_source=rsyncer_info.source, visit=visit_name).inc(
        rsyncer_info.increment_count
    )
    prom.seen_data_files.labels(rsync_source=rsyncer_info.source, visit=visit_name).inc(
        rsyncer_info.increment_data_count
    )


@router.post("/visits/{visit_name}/increment_rsync_transferred_files")
def increment_rsync_transferred_files(
    visit_name: str, rsyncer_info: RsyncerInfo, db=murfey_db
):
    rsync_instance = db.exec(
        select(RsyncInstance).where(
            RsyncInstance.source == rsyncer_info.source,
            RsyncInstance.destination == rsyncer_info.destination,
            RsyncInstance.session_id == rsyncer_info.session_id,
        )
    ).one()
    rsync_instance.files_transferred += rsyncer_info.increment_count
    db.add(rsync_instance)
    db.commit()
    db.close()


@router.post("/visits/{visit_name}/increment_rsync_transferred_files_prometheus")
def increment_rsync_transferred_files_prometheus(
    visit_name: str, rsyncer_info: RsyncerInfo, db=murfey_db
):
    prom.transferred_files.labels(
        rsync_source=rsyncer_info.source, visit=visit_name
    ).inc(rsyncer_info.increment_count)
    prom.transferred_files_bytes.labels(
        rsync_source=rsyncer_info.source, visit=visit_name
    ).inc(rsyncer_info.bytes)
    prom.transferred_data_files.labels(
        rsync_source=rsyncer_info.source, visit=visit_name
    ).inc(rsyncer_info.increment_data_count)
    prom.transferred_data_files_bytes.labels(
        rsync_source=rsyncer_info.source, visit=visit_name
    ).inc(rsyncer_info.data_bytes)


@router.post("/visits/{visit_name}/monitoring/{on}")
def change_monitoring_status(visit_name: str, on: int):
    prom.monitoring_switch.labels(visit=visit_name)
    prom.monitoring_switch.labels(visit=visit_name).set(on)


@router.get("/metrics/{metric_name}")
def inspect_prometheus_metrics(
    metric_name: str,
):
    """
    A debugging endpoint that returns the current contents of any Prometheus
    gauges and counters that have been set up thus far.
    """

    # Extract the Prometheus metric defined in the Prometheus module
    metric: Optional[Counter | Gauge] = getattr(prom, metric_name, None)
    if metric is None or not isinstance(metric, (Counter, Gauge)):
        raise LookupError("No matching metric was found")

    # Package contents into dict and return
    results = {}
    if hasattr(metric, "_metrics"):
        for i, (label_tuple, sub_metric) in enumerate(metric._metrics.items()):
            labels = dict(zip(metric._labelnames, label_tuple))
            labels["value"] = sub_metric._value.get()
            results[i] = labels
        return results
    else:
        value = metric._value.get()
        return {"value": value}



from logging import getLogger

from fastapi import APIRouter, Depends

from murfey.server.api.auth import MurfeySessionIDFrontend as MurfeySessionID
from murfey.server.api.auth import validate_token
from murfey.server.api.file_io_shared import GainReference
from murfey.server.api.file_io_shared import process_gain as _process_gain
from murfey.server.murfey_db import murfey_db

logger = getLogger("murfey.server.api.file_io_frontend")


router = APIRouter(
    prefix="/file_io/frontend",
    dependencies=[Depends(validate_token)],
    tags=["File I/O: Frontend"],
)


@router.post("/sessions/{session_id}/process_gain")
async def process_gain(
    session_id: MurfeySessionID, gain_reference_params: GainReference, db=murfey_db
):
    result = await _process_gain(session_id, gain_reference_params, db)
    return result



"""
API endpoints related to installing Murfey on a client.

Client machines may not have a direct internet connection, so Murfey allows
passing through download requests for Cygwin and MSYS2 to their respective
websites, and requests for Python packages to PyPI using the PEP 503 simple
API.

A static HTML page gives instructions on how to install on a network-isolated
system that has Python already installed. A previously set up system does not
need to have pip installed in order to bootstrap Murfey. Python and rsync are
required.
"""

from __future__ import annotations

import functools
import json
import logging
import random
import re
import zipfile
from io import BytesIO
from typing import Any
from urllib.parse import quote

import packaging.version
import requests
from fastapi import APIRouter, HTTPException, Query, Request, Response
from fastapi.responses import FileResponse, HTMLResponse, StreamingResponse

import murfey
from murfey.server.api import templates
from murfey.util.config import get_hostname, get_machine_config, get_microscope

tag = {
    "name": "Bootstrap",
    "description": __doc__,
    "externalDocs": {
        "description": "PEP 503",
        "url": "https://www.python.org/dev/peps/pep-0503/",
    },
}

# Set up API endpoint groups
# NOTE: Routers MUST have prefixes. prefix="" causes an error
version = APIRouter(prefix="/version", tags=["Bootstrap"])
bootstrap = APIRouter(prefix="/bootstrap", tags=["Bootstrap"])
cygwin = APIRouter(prefix="/cygwin", tags=["Bootstrap"])
msys2 = APIRouter(prefix="/msys2", tags=["Bootstrap"])
rust = APIRouter(prefix="/rust", tags=["Bootstrap"])
pypi = APIRouter(prefix="/pypi", tags=["Bootstrap"])
plugins = APIRouter(prefix="/plugins", tags=["Bootstrap"])

logger = logging.getLogger("murfey.server.api.bootstrap")

# Create a reusable HTTP session to avoid spamming external endpoints
http_session = requests.Session()

"""
=======================================================================================
GENERAL HELPER FUNCTIONS
=======================================================================================
"""


def _sanitise_str(input: str) -> str:
    # Remove \r and \n characters from the string
    input_clean = input.replace("\r", "").replace("\n", "").rstrip()
    return input_clean


def resolve_netloc(request: Request) -> str:
    """
    Helper function to construct the correct netloc (hostname[:port]) to use based on
    the request received. It will prioritise parsing the request headers for the host,
    port, protocol and using them to construct the netloc before defaulting to parsing
    the FastAPI Request object to do so.
    """

    # Prefer headers added by reverse proxies
    host = request.headers.get("X-Forwarded-Host", request.url.hostname)
    port = request.headers.get(
        "X-Forwarded-Port", str(request.url.port) if request.url.port else None
    )
    proto = request.headers.get("X-Forwarded-Proto", request.url.scheme)

    # Default ports shouldn't be included; if no ports are found, return just the host
    if (
        (proto == "http" and port == "80")
        or (proto == "https" and port == "443")
        or not port
    ):
        return host

    return f"{host}:{port}"


"""
=======================================================================================
VERSION-RELATED API ENDPOINTS
=======================================================================================
"""


@version.get("/")
def get_version(client_version: str = ""):
    result = {
        "server": murfey.__version__,
        "oldest-supported-client": murfey.__supported_client_version__,
    }

    if client_version:
        client = packaging.version.parse(client_version)
        server = packaging.version.parse(murfey.__version__)
        minimum_version = packaging.version.parse(murfey.__supported_client_version__)
        result["client-needs-update"] = minimum_version > client
        result["client-needs-downgrade"] = client > server

    return result


"""
=======================================================================================
GENERAL BOOTSTRAP-RELATED API ENDPOINTS
=======================================================================================
"""


def respond_with_template(
    request: Request, filename: str, parameters: dict[str, Any] | None = None
):
    template_parameters = {
        "hostname": get_hostname(),
        "microscope": get_microscope(),
        "version": murfey.__version__,
        # Extra parameters to reconstruct URLs for forwarded requests
        "netloc": request.url.netloc,
        "proxy_path": "",
    }
    if parameters:
        template_parameters.update(parameters)
    return templates.TemplateResponse(
        request=request, name=filename, context=template_parameters
    )


@bootstrap.get("/", response_class=HTMLResponse)
def get_bootstrap_instructions(request: Request):
    """
    Return a website containing instructions for installing the Murfey client on a
    machine with no internet access.
    """

    # Check if this is a forwarded request from somewhere else and construct netloc
    netloc = resolve_netloc(request)

    # Find path to 'bootstrap' router using current URL path
    proxy_path = request.url.path.removesuffix(f"{bootstrap.prefix}/")

    return respond_with_template(
        request=request,
        parameters={
            "netloc": netloc,
            "proxy_path": proxy_path,
        },
        filename="bootstrap.html",
    )


@bootstrap.get("/pip.whl", response_class=StreamingResponse)
def get_pip_wheel():
    """
    Return a static version of pip. This does not need to be the newest or best,
    but has to be compatible with all supported Python versions.
    This is only used during bootstrapping by the client to identify and then
    download the actually newest appropriate version of pip.
    """
    return get_pypi_file(
        package="pip",
        filename="pip-22.2.2-py3-none-any.whl",  # Highest version that works before PEP 658 change
    )


@bootstrap.get("/murfey.whl", response_class=StreamingResponse)
def get_murfey_wheel():
    """
    Return a wheel file containing the latest release version of Murfey. We should
    not have to worry about the exact Python compatibility here, as long as
    murfey.bootstrap is compatible with all relevant versions of Python.
    This also ignores yanked releases, which again should be fine.
    """
    full_path_response = http_session.get(f"{pypi_index_url.rstrip('/')}/murfey")
    wheels = {}

    for wheel_file in re.findall(
        b"<a [^>]*>([^<]*).whl</a>",
        full_path_response.content,
    ):
        try:
            filename = wheel_file.decode("utf-8") + ".whl"
            version = packaging.version.parse(filename.split("-")[1])
            wheels[version] = filename
        except Exception:
            pass  # Ignore searches that fail to return wheels
    if not wheels:
        raise HTTPException(
            status_code=404, detail="Could not identify appropriate version of Murfey"
        )
    newest_version = max(wheels)
    return get_pypi_file(package="murfey", filename=wheels[newest_version])


"""
=======================================================================================
CYGWIN-RELATED API ENDPOINTS
=======================================================================================
"""


@cygwin.get("/setup-x86_64.exe", response_class=StreamingResponse)
def get_cygwin_setup():
    """
    Obtain and pass through a Cygwin installer from an official source.
    This is used during client bootstrapping and can download and install the
    Cygwin distribution that then remains on the client machines.
    """
    filename = "setup-x86_64.exe"
    response = http_session.get(f"https://www.cygwin.com/{filename}")

    # Construct headers to return with response
    headers: dict[str, str] = {
        "Content-Disposition": f"attachment; filename=cygwin-{filename}"
    }
    if response.headers.get("Content-Length"):
        headers["Content-Length"] = response.headers["Content-Length"]

    return StreamingResponse(
        content=response.iter_content(chunk_size=8192),
        status_code=response.status_code,
        headers=headers,
        media_type=response.headers.get("Content-Type"),
    )


@functools.lru_cache()
def find_cygwin_mirror() -> str:
    """
    Find an appropriate Cygwin mirror to pass download requests through to.
    We don't expect these to change often, so we only do this once during the
    lifetime of the server.
    """
    url = "https://www.cygwin.com/mirrors.lst"
    mirrors = http_session.get(url)
    logger.info(
        f"Reading mirrors from {url} returned status code {mirrors.status_code} {mirrors.reason}"
    )

    # Don't cache result if we can't get mirrors list
    assert mirrors.status_code == 200

    mirror_priorities = {}
    for mirror in mirrors.content.split(b"\n"):
        mirror_line = mirror.decode("utf-8").strip().split(";")
        if not mirror_line or len(mirror_line) < 4:
            continue
        if not mirror_line[0].startswith("http"):
            continue
        if mirror_line[3] == "UK":
            priority = 20
        elif mirror_line[2] == "Europe":
            priority = 10
        else:
            priority = 0
        if mirror_line[0].startswith("https"):
            priority += 5
        mirror_priorities[mirror_line[0]] = priority

    elegible_mirrors = [
        mirror
        for mirror, priority in mirror_priorities.items()
        if priority == max(mirror_priorities.values())
    ]
    if not elegible_mirrors:
        logger.warning("No valid mirrors identified")
        assert elegible_mirrors

    picked_mirror = random.choice(elegible_mirrors)
    if not picked_mirror.endswith("/"):
        picked_mirror += "/"
    logger.info(f"Picked Cygwin mirror: {picked_mirror}")
    return picked_mirror


@cygwin.get("/{request_path:path}", response_class=StreamingResponse)
def parse_cygwin_request(
    request: Request,
    request_path: str,
):
    """
    Forward a Cygwin setup request to an official mirror.
    """

    logger.debug(f"Received request to access {str(request.url)!r}")

    # Validate request path
    if bool(re.fullmatch(r"^[\w\s\.\-\+/]+$", request_path)) is False:
        raise ValueError(f"{request_path!r} is not a valid request path")

    try:
        url = f'{find_cygwin_mirror()}{quote(request_path, safe="/")}'
    except Exception:
        raise HTTPException(
            status_code=503, detail="Could not identify a suitable Cygwin mirror"
        )

    logger.info(f"Forwarding Cygwin download request to {_sanitise_str(url)}")
    response = http_session.get(url)

    headers: dict[str, str] = {}
    if response.headers.get("Content-Length"):
        headers["Content-Length"] = response.headers["Content-Length"]

    return StreamingResponse(
        content=response.iter_content(chunk_size=8192),
        status_code=response.status_code,
        headers=headers,
        media_type=response.headers.get("Content-Type"),
    )


"""
=======================================================================================
MSYS2-RELATED FUNCTIONS AND ENDPOINTS
=======================================================================================
"""

# Variables used by the MSYS2 functions below
msys2_url = "https://repo.msys2.org"
msys2_file_ext = (".exe", ".sig", ".tar.xz", "tar.zst")
valid_envs = (
    # Tuple of systems and supported libraries/compilers/architectures within
    (
        "msys",  # Cygwin-like system
        (
            # Available environments
            "i686",  # 32-bit
            "x86_64",  # 64-bit
            "sources",
        ),
    ),
    (
        "mingw",  # Windows-like system
        (
            # Available environments
            # Toolchain: llvm, C library:   ucrt, C++ library:    libc++
            "clang32",
            "clang64",
            "clangarm64",
            # Toolchain:  gcc, C library:   ucrt, C++ library: libstdc++
            "ucrt64",
            # Toolchain:  gcc, C library: msvcrt, C++ library: libstdc++
            "mingw32",
            "mingw64",
            # Architecture types
            "i686",  # 32-bit
            "x86_64",  # 64-bit
            "sources",
        ),
    ),
)


@msys2.get("/config/pacman.d.zip", response_class=StreamingResponse)
def get_pacman_mirrors(request: Request):
    """
    Dynamically generates a zip file containing mirrorlist files that have been set
    up to mirror the MSYS2 package database for each environment.

    The files in this folder should be pasted into, and overwrite, the 'mirrorlist'
    files present in the %MSYS64%\\etc\\pacman.d folder. The default path to this
    folder is C:\\msys64\\etc\\pacman.d.
    """

    # Check if this is a forwarded request from somewhere else and construct netloc
    netloc = resolve_netloc(request)

    # Find path to Rust router using current URL Path
    path_to_router = request.url.path.removesuffix("/config/pacman.d.zip")

    # Construct base URL for subsequent use
    base_url = f"{request.url.scheme}://{netloc}{path_to_router}"
    logger.debug(f"Base URL to MSYS2 sub-router determined to be {base_url}")

    # Construct package database mirrors
    # Files are called mirrorlist.{environment}
    # URL format: {scheme}://{netloc}{proxy_path}/{router_prefix}/path/to/repo
    url_paths = {
        "clang64": "mingw/clang64",
        "mingw": "mingw/$repo",
        "mingw32": "mingw/i686",
        "mingw64": "mingw/x86_64",
        "msys": "msys/$arch",
        "ucrt64": "mingw/ucrt64",
    }
    # Construct file names and contents
    mirror_lists = {
        f"mirrorlist.{env}": "\n".join(
            [
                "# See https://www.msys2.org/dev/mirrors",
                "",
                "## Primary",
                f"Server = {base_url}/repo/{repo_path}",
                "",
            ]
        )
        for env, repo_path in url_paths.items()
    }

    # Create in-memory buffer for the ZIP file
    zip_buffer = BytesIO()

    # Create a zip file in the buffer
    with zipfile.ZipFile(zip_buffer, "w", zipfile.ZIP_DEFLATED) as zip_file:
        for file_name, content in mirror_lists.items():
            zip_file.writestr(file_name, content)
    zip_buffer.seek(0)  # Move object pointer back to start

    # Construct and return streaming response
    headers = {
        "Content-Disposition": "attachment; filename=pacman.d.zip",
        "Content-Length": str(zip_buffer.getbuffer().nbytes),
    }
    return StreamingResponse(
        zip_buffer,
        status_code=200,
        headers=headers,
        media_type="application/zip",
    )


@msys2.get("/repo/distrib/{setup_file}", response_class=StreamingResponse)
def get_msys2_setup(
    request: Request,
    setup_file: str,
):
    """
    Obtain and pass through an MSYS2 installer from an official source.
    This is used during client bootstrapping, and can download and install the
    MSYS2 distribution that then remains on the client machines.
    """

    logger.debug(f"Received request to access {str(request.url)!r}")

    # Validate characters in sent path
    if not bool(re.fullmatch(r"^[\w\.\-]+$", setup_file)):
        raise ValueError("Unallowed characters present in requested setup file")

    # Allow only '.exe', 'tar.xz', 'tar.zst', or '.sig' files
    if not setup_file.startswith("msys2") and not any(
        setup_file.endswith(ext) for ext in (msys2_file_ext)
    ):
        raise ValueError(f"{setup_file!r} is not a valid executable")

    response = http_session.get(f"{msys2_url}/distrib/{setup_file}")

    headers: dict[str, str] = {}
    if response.headers.get("Content-Length"):
        headers["Content-Length"] = response.headers["Content-Length"]

    return StreamingResponse(
        content=response.iter_content(chunk_size=8192),
        status_code=response.status_code,
        headers=headers,
        media_type=response.headers.get("Content-Type"),
    )


@msys2.get("/repo/", response_class=Response)
def get_msys2_main_index(
    request: Request,
) -> Response:
    """
    Returns a simple index displaying valid MSYS2 systems and the latest setup file
    from the main MSYS2 repository.
    """

    logger.debug(f"Received request to access {str(request.url)!r}")

    # Construct URL and get response
    env_url = f"{msys2_url}"
    response = http_session.get(env_url)

    # Parse and rewrite package index content
    content: bytes = response.content  # Get content in bytes
    content_text: str = content.decode("utf-8")  # Convert to strings
    content_text_list = []
    for line in content_text.splitlines():
        if line.startswith("<a href"):
            # Mirror only lines related to MSYS2 environments or the distribution folder
            if any(env[0] in line for env in valid_envs) or "distrib" in line:
                content_text_list.append(line)
            # Other URLs don't need to be mirrored
            else:
                continue
        else:
            content_text_list.append(line)

    # Reconstruct conent
    content_text_new = str("\n".join(content_text_list))  # Regenerate HTML structure
    content_new = content_text_new.encode("utf-8")  # Convert back to bytes
    return Response(
        content=content_new,
        status_code=response.status_code,
        media_type=response.headers.get("Content-Type"),
    )


@msys2.get("/repo/{system}/", response_class=Response)
def get_msys2_environment_index(
    request: Request,
    system: str,
) -> Response:
    """
    Returns a list of all MSYS2 environments for a given system from the main MSYS2
    repository.
    """

    logger.debug(f"Received request to access {str(request.url)!r}")

    # Validate provided system; use this endpoint to display 'distrib' folder too
    if not (any(system in env[0] for env in valid_envs) or system == "distrib"):
        raise ValueError(f"{system!r} is not a valid msys2 environment")

    # Construct URL to main MSYS repo and get response
    arch_url = f'{msys2_url}/{quote(system, safe="/")}'
    response = http_session.get(arch_url)

    # Parse and rewrite package index content
    content: bytes = response.content  # Get content in bytes
    content_text: str = content.decode("utf-8")  # Convert to strings
    content_text_list = []
    for line in content_text.splitlines():
        if line.startswith("<a href="):
            # Skip non-executable files when querying 'distrib' repo
            if system == "distrib":
                if not any(ext in line for ext in msys2_file_ext):
                    continue
            content_text_list.append(line)
        else:
            content_text_list.append(line)

    # Reconstruct conent
    content_text_new = str("\n".join(content_text_list))  # Regenerate HTML structure
    content_new = content_text_new.encode("utf-8")  # Convert back to bytes
    return Response(
        content=content_new,
        status_code=response.status_code,
        media_type=response.headers.get("Content-Type"),
    )


@msys2.get("/repo/{system}/{environment}/", response_class=Response)
def get_msys2_package_index(
    request: Request,
    system: str,
    environment: str,
) -> Response:
    """
    Obtain a list of all available MSYS2 packages for a given environment from the main
    MSYS2 repo.
    """

    logger.debug(f"Received request to access {str(request.url)!r}")

    # Validate environment
    if any(system in env[0] and environment in env[1] for env in valid_envs) is False:
        raise ValueError(f"{system!r}/{environment!r} is not a valid msys2 environment")

    # Construct URL to main MSYS repo and get response
    package_list_url = (
        f'{msys2_url}/{quote(system, safe="/")}/{quote(environment, safe="/")}'
    )
    response = http_session.get(package_list_url)
    return Response(
        content=response.content,
        status_code=response.status_code,
        media_type=response.headers.get("Content-Type"),
    )


@msys2.get("/repo/{system}/{environment}/{package}", response_class=StreamingResponse)
def get_msys2_package_file(
    request: Request,
    system: str,
    environment: str,
    package: str,
) -> Response:
    """
    Obtain and pass through a specific download for an MSYS2 package.
    """

    logger.debug(f"Received request to access {str(request.url)!r}")

    # Validate environment
    if any(system in env[0] and environment in env[1] for env in valid_envs) is False:
        raise ValueError(f"'{system}/{environment}' is not a valid msys2 environment")

    # Validate package name
    #   MSYS2 package names contain:
    #   - alphanumerics (includes "_"; \w),
    #   - periods (\.),
    #   - dashes (\-),
    #   - tildes (~), and
    #   - plus signs (+)
    if bool(re.fullmatch(r"^[\w\.\-\+~]+$", package)) is False:
        raise ValueError(f"{package!r} is not a valid package name")

    # Construct URL to main MSYS repo and get response
    package_url = f'{msys2_url}/{quote(system, safe="/")}/{quote(environment, safe="/")}/{quote(package, safe="/")}'
    response = http_session.get(package_url)
    if response.status_code != 200:
        raise HTTPException(status_code=response.status_code)

    headers: dict[str, str] = {}
    if response.headers.get("Content-Length"):
        headers["Content-Length"] = response.headers["Content-Length"]

    return StreamingResponse(
        content=response.iter_content(chunk_size=8192),
        status_code=response.status_code,
        headers=headers,
        media_type=response.headers.get("Content-Type"),
    )


"""
=======================================================================================
RUST-RELATED FUNCTIONS AND ENDPOINTS
=======================================================================================
"""

# Base URLs to use
rust_index = "https://index.crates.io"
rust_dl = "https://static.crates.io/crates"
rust_api = "https://crates.io"


@rust.get("/cargo/config.toml", response_class=StreamingResponse)
def get_cargo_config(request: Request):
    """
    Returns a properly configured Cargo config that sets it to look ONLY at the
    crates.io mirror.

    The default path for this config on Linux devices is ~/.cargo/config.toml,
    and its default path on Windows is %USERPROFILE%\\.cargo\\config.toml.
    """

    # Check if this is a forwarded request from somewhere else and construct netloc
    netloc = resolve_netloc(request)

    # Find path to Rust router using current URL Path
    path_to_router = request.url.path.removesuffix("/cargo/config.toml")

    # Construct base URL for subsequent use
    base_url = f"{request.url.scheme}://{netloc}{path_to_router}"
    logger.debug(f"Base URL to Rust sub-router determined to be {base_url}")

    # Construct URL to our mirror of the Rust sparse index
    index_url = f"{base_url}/index/"

    # Construct and return the config.toml file
    config_data = "\n".join(
        [
            "[source.crates-io]",
            'replace-with = "murfey-crates"',  # Redirect to our mirror
            "",
            "[source.murfey-crates]",
            f'registry = "sparse+{index_url}"',  # sparse+ to use sparse protocol
            "",
            "[registries.murfey-crates]",
            f'index = "sparse+{index_url}"',  # sparse+ to use sparse protocol
            "",
            "[registry]",
            'default = "murfey-crates"',  # Redirect to our mirror
            "",
        ]
    )
    config_bytes = config_data.encode("utf-8")

    headers: dict[str, str] = {
        "Content-Disposition": "attachment; filename=config.toml",
        "Content-Length": str(len(config_bytes)),
    }

    return StreamingResponse(
        BytesIO(config_bytes),
        status_code=200,
        headers=headers,
        media_type="application/toml+json",
    )


"""
crates.io Sparse Index Registry Key Endpoints
"""


@rust.get("/index/", response_class=Response)
def get_index_page():
    """
    Returns a mirror of the https://index.crates.io landing page.
    """

    response = http_session.get(rust_index)
    if response.status_code != 200:
        raise HTTPException(status_code=response.status_code)
    return Response(
        content=response.content,
        status_code=response.status_code,
        media_type=response.headers.get("Content-Type"),
    )


@rust.get("/index/config.json", response_class=StreamingResponse)
def get_index_config(request: Request):
    """
    Download a config.json file used by Cargo to navigate sparse index registries
    with.

    The 'dl' key points to our mirror of the static crates.io repository, while
    the 'api' key points to an API version of that same registry. Both will be
    used by Cargo when searching for and downloading packages.
    """

    # Check if this is a forwarded request from somewhere else and construct netloc
    netloc = resolve_netloc(request)

    # Find path to Rust router using current URL Path
    path_to_router = request.url.path.removesuffix("/index/config.json")

    # Construct base URL for subsequent use
    base_url = f"{request.url.scheme}://{netloc}{path_to_router}"
    logger.debug(f"Base URL to Rust sub-router determined to be {base_url}")

    # Construct config file with the necessary endpoints
    config = {
        "dl": f"{base_url}/crates",
        "api": f"{base_url}",
    }

    # Save it as a JSON and return it as part of the response
    json_data = json.dumps(config, indent=4)
    json_bytes = json_data.encode("utf-8")

    headers: dict[str, str] = {
        "Content-Disposition": "attachment; filename=config.json",
        "Content-Length": str(len(json_bytes)),
    }

    return StreamingResponse(
        BytesIO(json_bytes),
        status_code=200,
        headers=headers,
        media_type="application/json",
    )


@rust.get("/index/{c1}/{c2}/{package}", response_class=StreamingResponse)
def get_index_package_metadata(
    request: Request,
    c1: str,
    c2: str,
    package: str,
):
    """
    Download the metadata for a given package from the crates.io sparse index.
    The path to the metadata file on the server side takes the following form:
    /{c1}/{c2}/{package}

    c1 and c2 are 2 characters-long strings that are taken from the first 4
    characters of the package name (a-z, A-Z, 0-9, -, _). For 3-letter packages,
    c1 = 3, and c2 is the first character of the package.
    """

    logger.debug(f"Received request to access {str(request.url)!r}")

    # Validate path to the package metadata
    if any(not re.fullmatch(r"[\w\-]{1,2}", char) for char in (c1, c2)):
        raise ValueError("Invalid path to package metadata")

    if len(c1) == 1 and not c1 == "3":
        raise ValueError("Invalid path to package metadata")
    if c1 == "3" and not len(c2) == 1:
        raise ValueError("Invalid path to package metadata")

    if not re.fullmatch(r"[\w\-]+", package):
        raise ValueError("Invalid package name")

    # Request and return the metadata as a JSON file
    url = f"{rust_index}/{c1}/{c2}/{package}"
    response = http_session.get(url)
    if response.status_code != 200:
        raise HTTPException(status_code=response.status_code)
    return StreamingResponse(
        response.iter_content(chunk_size=8192),
        status_code=response.status_code,
        media_type=response.headers.get("Content-Type"),
    )


@rust.get("/index/{n}/{package}", response_class=StreamingResponse)
def get_index_package_metadata_for_short_package_names(
    request: Request,
    n: str,
    package: str,
):
    """
    The Rust sparse index' naming scheme for packages with 1-2 characters is
    different from the standard path convention. They are stored under
    /1/{package} or /2/{package}.
    """

    logger.debug(f"Received request to access {str(request.url)!r}")

    # Validate path to crate
    if n not in ("1", "2"):
        raise ValueError("Invalid path to package metadata")
    if not re.fullmatch(r"[\w\-]{1,2}", package):
        raise ValueError("Invalid package name")

    # Request and return the metadata as a JSON file
    url = f"{rust_index}/{n}/{package}"
    response = http_session.get(url)
    if response.status_code != 200:
        raise HTTPException(status_code=response.status_code)
    return StreamingResponse(
        response.iter_content(chunk_size=8192),
        status_code=response.status_code,
        media_type=response.headers.get("Content-Type"),
    )


@rust.get("/crates/{package}/{version}/download", response_class=StreamingResponse)
def get_rust_package_download(
    request: Request,
    package: str,
    version: str,
):
    """
    Obtain and pass through a crate download request for a Rust package via the
    sparse index registry.
    """

    logger.debug(f"Received request to access {str(request.url)!r}")

    # Validate package and version
    if not re.fullmatch(r"[\w\-]+", package):
        raise ValueError("Invalid package name")
    if not re.fullmatch(r"[\w\-\.\+]+", version):
        raise ValueError("Invalid version number")

    # Request and return crate from https://static.crates.io
    url = f"{rust_dl}/{package}/{version}/download"
    response = http_session.get(url)
    file_name = f"{package}-{version}.crate"  # Construct file name to save package as
    if response.status_code != 200:
        raise HTTPException(status_code=response.status_code)

    headers = {"Content-Disposition": f'attachment; filename="{file_name}"'}
    if response.headers.get("Content-Length"):
        headers["Content-Length"] = response.headers["Content-Length"]

    return StreamingResponse(
        content=response.iter_content(chunk_size=8192),
        status_code=response.status_code,
        headers=headers,
        media_type=response.headers.get("Content-Type", "application/octet-stream"),
    )


"""
crates.io API Key Endpoints
"""


@rust.get("/api/v1/crates")
def get_rust_api_package_index(
    request: Request,
    package: str = Query(None, alias="q"),
    per_page: int = Query(10),
    cursor: str = Query(None, alias="seek"),
):
    """
    Displays the Rust API package index, which returns names of available packages
    in a JSON object based on the search query given.
    """

    logger.debug(f"Received request to access {str(request.url)!r}")

    # Validate package name
    if package and not re.fullmatch(r"[\w\-]+", package):
        raise ValueError("Invalid package name")
    if cursor and not re.fullmatch(r"[a-zA-Z0-9]+", cursor):
        raise ValueError("Invalid cursor")

    # Formulate the search query to pass to the crates page
    search_params: dict[str, str | int] = {}
    if package:
        search_params["q"] = package
    if per_page:
        search_params["per_page"] = per_page
    if cursor:
        search_params["seek"] = cursor

    # Submit request and return response
    url = f"{rust_api}/api/v1/crates"
    response = http_session.get(url, params=search_params)
    if response.status_code != 200:
        raise HTTPException(status_code=response.status_code)
    return response.json()


@rust.get("/api/v1/crates/{package}")
def get_rust_api_package_info(
    request: Request,
    package: str,
):
    """
    Displays general information for a given Rust package, as a JSON object.
    Contains both version information and download information, in addition
    to other types of metadata.
    """

    logger.debug(f"Received request to access {str(request.url)!r}")

    # Validate package name
    if not re.fullmatch(r"[\w\-]+", package):
        raise ValueError("Invalid package name")

    # Return JSON of the package's page
    url = f"{rust_api}/api/v1/crates/{package}"
    response = http_session.get(url)
    if response.status_code != 200:
        raise HTTPException(status_code=response.status_code)
    return response.json()


@rust.get("/api/v1/crates/{package}/versions")
def get_rust_api_package_versions(
    request: Request,
    package: str,
):
    """
    Displays all available versions for a particular Rust package, along with download
    links for said versions, as a JSON object.
    """

    logger.debug(f"Received request to access {str(request.url)!r}")

    # Validate crate name
    if not re.fullmatch(r"[\w\-]+", package):
        raise ValueError("Invalid package name")

    # Return JSON of the package's version information
    url = f"{rust_api}/api/v1/crates/{package}/versions"
    response = http_session.get(url)
    if response.status_code != 200:
        raise HTTPException(status_code=response.status_code)
    return response.json()


@rust.get(
    "/api/v1/crates/{package}/{version}/download", response_class=StreamingResponse
)
def get_rust_api_package_download(
    request: Request,
    package: str,
    version: str,
):
    """
    Obtain and pass through a crate download request for a specific Rust package.
    """

    logger.debug(f"Received request to access {str(request.url)!r}")

    # Validate package name
    if not re.fullmatch(r"[\w\-]+", package):
        raise ValueError("Invalid package name")
    # Validate version number
    # Not all developers adhere to guidelines when versioning their packages, so
    # '-', '_', '+', as well as letters can also be present in this field.
    if not re.fullmatch(r"[\w\-\.\+]+", version):
        raise ValueError("Invalid version number")

    # Request and return package
    url = f"{rust_api}/api/v1/crates/{package}/{version}/download"
    response = http_session.get(url)
    file_name = f"{package}-{version}.crate"  # Construct crate name to save as
    if response.status_code != 200:
        raise HTTPException(status_code=response.status_code)

    headers = {"Content-Disposition": f'attachment; filename="{file_name}"'}
    if response.headers.get("Content-Length"):
        headers["Content-Length"] = response.headers["Content-Length"]

    return StreamingResponse(
        content=response.iter_content(chunk_size=8192),
        status_code=response.status_code,
        headers=headers,
        media_type=response.headers.get("Content-Type", "application/octet-stream"),
    )


@rust.get("/crates/{package}/{crate}", response_class=StreamingResponse)
def get_rust_package_crate(
    request: Request,
    package: str,
    crate: str,
):
    """
    Obtain and pass through a download for a specific Rust crate. The Rust API
    download request actually redirects to the static crate repository, so this
    endpoint covers cases where the static crate download link is requested.

    The static Rust package repository has been configured such that only requests
    for a specific crate are accepted and handled.
    (e.g. https://static.crates.io/crates/anyhow/anyhow-1.0.97.crate will pass)

    A request for any other part of the URL path will be denied.
    (e.g. https://static.crates.io/crates/anyhow will fail)
    """

    logger.debug(f"Received request to access {str(request.url)!r}")

    # Validate crate and package names
    if not re.fullmatch(r"[\w\-]+", package):
        raise ValueError("Invalid package name")
    if not crate.endswith(".crate"):
        raise ValueError("This is a not a Rust crate")
    # Rust crates follow a '{crate}-{version}.crate' structure
    if not re.fullmatch(r"[\w\-\.]+\.crate", crate):
        raise ValueError("Invalid crate name")

    # Request and return package
    url = f"{rust_dl}/{package}/{crate}"
    response = http_session.get(url)
    if response.status_code != 200:
        raise HTTPException(status_code=response.status_code)

    headers = {"Content-Disposition": f'attachment; filename="{crate}"'}
    if response.headers.get("Content-Length"):
        headers["Content-Length"] = response.headers["Content-Length"]

    return StreamingResponse(
        content=response.iter_content(chunk_size=8192),
        status_code=response.status_code,
        headers=headers,
        media_type=response.headers.get("Content-Type", "application/octet-stream"),
    )


"""
=======================================================================================
PYPI-RELATED FUNCTIONS AND ENDPOINTS
=======================================================================================
"""

python_repo_url = "https://files.pythonhosted.org"
pypi_index_url = "https://pypi.org/simple/"


def _get_full_pypi_path_response(package: str) -> requests.Response:
    """
    Validates the package name, sanitises it if valid, and attempts to return a HTTP
    response from PyPI.
    """

    # Check that a package name follows PEP 503 naming conventions, containing only
    # alphanumerics (including underscores; \w), dashes (\-), and periods (\.)
    if not re.fullmatch(r"[\w\-\.]+", package):
        raise ValueError(f"{package!r} is not a valid package name")

    # Sanitise and normalise package name according to PEP 503
    package_clean = quote(re.sub(r"[-_.]+", "-", package.lower()), safe="/")

    # Get HTTP response
    url = f"{pypi_index_url.rstrip('/')}/{package_clean}"
    response = http_session.get(url)
    if response.status_code != 200:
        raise HTTPException(status_code=response.status_code)
    return response


@pypi.get("/index/", response_class=Response)
def get_pypi_index():
    """
    Obtain list of all PyPI packages via the simple API (PEP 503).
    """

    response = http_session.get(pypi_index_url)
    return Response(
        content=response.content,
        status_code=response.status_code,
        media_type=response.headers.get("Content-Type"),
    )


@pypi.get("/index/{package}/", response_class=Response)
def get_pypi_package_downloads_list(request: Request, package: str) -> Response:
    """
    Obtain list of all package downloads from PyPI via the simple API (PEP 503), and
    rewrite all download URLs to point to this server, under the current directory.
    """

    logger.debug(f"Received request to access {str(request.url)!r}")

    # Construct base URL to rewrite with
    netloc = resolve_netloc(request)
    scheme = request.headers.get("X-Forwarded-Proto", request.url.scheme)
    router_path = request.url.path.removesuffix(f"/index/{package}/")
    base_url = f"{scheme}://{netloc}{router_path}"

    # Validate package and URL
    full_path_response = _get_full_pypi_path_response(package)

    # Process lines related to PyPI packages in response
    content: bytes = full_path_response.content  # In bytes
    content_text: str = content.decode("utf-8")  # Convert to strings

    # PyPI's simple index now directly points to https://pythonhosted.org
    # It also uses newlines partway through the '<a ...></a>' blocks now
    # It's thus now better to use regex substitution on the page as a whole
    content_text_new = re.sub(re.escape(python_repo_url), base_url, content_text)

    content_new = content_text_new.encode("utf-8")  # Convert back to bytes

    return Response(
        content=content_new,
        status_code=full_path_response.status_code,
        media_type=full_path_response.headers.get("Content-Type"),
    )


@pypi.get("/packages/{a}/{b}/{c}/{filename}", response_class=StreamingResponse)
def get_pypi_file(
    request: Request,
    a: str,
    b: str,
    c: str,
    filename: str,
):
    """
    Obtain and pass through a specific download for a PyPI package.
    """
    logger.debug(f"Received request to access {str(request.url)!r}")

    package_url = f"{python_repo_url}/packages/{a}/{b}/{c}/{filename}"
    logger.debug(f"Forwarding package request to {package_url!r}")
    response = http_session.get(package_url, stream=True)

    # Construct headers to return with response
    headers: dict[str, str] = {}
    if response.status_code != 200:
        raise HTTPException(status_code=response.status_code)
    return StreamingResponse(
        content=response.raw,
        status_code=response.status_code,
        headers=headers,
        media_type=response.headers.get("Content-Type"),
    )


"""
=======================================================================================
PYPI API ENDPOINT PLUGINS
=======================================================================================
"""


@plugins.get("/instruments/{instrument_name}/{package}", response_class=FileResponse)
def get_plugin_wheel(instrument_name: str, package: str):
    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]
    wheel_path = machine_config.plugin_packages.get(package)
    if wheel_path is None:
        return None
    return FileResponse(
        wheel_path,
        headers={"Content-Disposition": "attachment; filename={wheel_path.name}"},
    )



from __future__ import annotations

import secrets
import time
from logging import getLogger
from typing import Dict
from uuid import uuid4

import aiohttp
import requests
from fastapi import APIRouter, Depends, HTTPException, status
from fastapi.security import (
    APIKeyCookie,
    OAuth2PasswordBearer,
    OAuth2PasswordRequestForm,
)
from jose import JWTError, jwt
from passlib.context import CryptContext
from pydantic import BaseModel
from sqlmodel import Session, create_engine, select
from typing_extensions import Annotated

from murfey.server.murfey_db import murfey_db, url
from murfey.util.api import url_path_for
from murfey.util.config import get_security_config
from murfey.util.db import MurfeyUser as User
from murfey.util.db import Session as MurfeySession

# Set up logger
logger = getLogger("murfey.server.api.auth")

# Set up router
router = APIRouter(
    prefix="/auth",
    tags=["Authentication"],
)


# Set up variables used for authentication
security_config = get_security_config()
auth_url = security_config.auth_url
ALGORITHM = security_config.auth_algorithm or "HS256"
SECRET_KEY = security_config.auth_key or secrets.token_hex(32)
if security_config.auth_type == "password":
    oauth2_scheme = OAuth2PasswordBearer(tokenUrl="token")
else:
    oauth2_scheme = APIKeyCookie(name=security_config.cookie_key)
if security_config.instrument_auth_type == "token":
    instrument_oauth2_scheme = OAuth2PasswordBearer(tokenUrl="token")
else:
    instrument_oauth2_scheme = lambda *args, **kwargs: None
pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")

instrument_server_tokens: Dict[float, dict] = {}

# Set up database engine
try:
    _url = url(security_config)
    engine = create_engine(_url)
except Exception:
    engine = None


def hash_password(password: str) -> str:
    return pwd_context.hash(password)


"""
=======================================================================================
TOKEN VALIDATION FUNCTIONS
=======================================================================================

Functions and helpers used to validate incoming requests from both the client and
the frontend. 'validate_token()' and 'validate_instrument_token()' are imported
int the other FastAPI modules and attached as dependencies to the routers.
"""


def check_user(username: str) -> bool:
    try:
        with Session(engine) as murfey_db:
            users = murfey_db.exec(select(User)).all()
    except Exception:
        return False
    return username in [u.username for u in users]


async def validate_token(token: Annotated[str, Depends(oauth2_scheme)]):
    """
    Used by the backend routers to validate requests coming in from frontend.
    """
    try:
        # Validate using auth URL if provided; will error if invalid
        if auth_url:
            headers = (
                {}
                if security_config.auth_type == "cookie"
                else {"Authorization": f"Bearer {token}"}
            )
            cookies = (
                {security_config.cookie_key: token}
                if security_config.auth_type == "cookie"
                else {}
            )
            async with aiohttp.ClientSession(cookies=cookies) as session:
                async with session.get(
                    f"{auth_url}/validate_token",
                    headers=headers,
                ) as response:
                    success = response.status == 200
                    validation_outcome = await response.json()
            if not (success and validation_outcome.get("valid")):
                raise JWTError
        # If authenticating using cookies; an auth URL MUST be provided
        else:
            if security_config.auth_type == "cookie":
                raise JWTError
        # Validate using password
        if security_config.auth_type == "password":
            decoded_data = jwt.decode(token, SECRET_KEY, algorithms=[ALGORITHM])
            # Check that the user is present and is valid
            if decoded_data.get("user"):
                if not check_user(decoded_data["user"]):
                    raise JWTError
            else:
                raise JWTError
    except JWTError:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Could not validate credentials from frontend",
            headers={"WWW-Authenticate": "Bearer"},
        )
    return None


def validate_session_against_visit(session_id: int, visit: str):
    """
    Checks that the session ID is associated with the claimed visit.
    """
    with Session(engine) as murfey_db:
        session_data = murfey_db.exec(
            select(MurfeySession).where(MurfeySession.id == session_id)
        ).all()
    if len(session_data) != 1:
        return False
    return visit == session_data[0].visit


async def validate_instrument_token(
    token: Annotated[str, Depends(instrument_oauth2_scheme)]
):
    """
    Used by the backend routers to check the incoming instrument server token.
    """
    try:
        # Validate using auth URL if provided
        if security_config.instrument_auth_url:
            async with aiohttp.ClientSession() as session:
                headers = (
                    {}
                    if not security_config.instrument_auth_type
                    else {"Authorization": f"Bearer {token}"}
                )
                async with session.get(
                    f"{security_config.instrument_auth_url}/validate_token",
                    headers=headers,
                ) as response:
                    success = response.status == 200
                    validation_outcome = await response.json()
            if not (success and validation_outcome.get("valid")):
                raise JWTError
        else:
            # First, check if the token has expired
            decoded_data = jwt.decode(token, SECRET_KEY, algorithms=[ALGORITHM])
            if expiry_time := decoded_data.get("expiry_time"):
                if expiry_time < time.time():
                    raise JWTError
            # Check that the decoded session corresponds to the visit
            elif decoded_data.get("session") is not None:
                if not validate_session_against_visit(
                    decoded_data["session"], decoded_data["visit"]
                ):
                    raise JWTError
            # Verify 'user' token if enabled
            elif security_config.allow_user_token:
                if not decoded_data.get("user"):
                    raise JWTError
            else:
                raise JWTError
    except JWTError:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Could not validate credentials from instrument",
            headers={"WWW-Authenticate": "Bearer"},
        )
    return None


"""
=======================================================================================
SESSION ID VALIDATION
=======================================================================================

Annotated ints are defined here that trigger validation of the session IDs in incoming
requests, verifying that the session is allowed to access the particular visit.

The 'MurfeySessionID...' types are imported and used in the type hints of the endpoint
functions in the other FastAPI routers, depending on whether requests from the frontend
or the instrument are expected.
"""


def get_visit_name(session_id: int) -> str:
    with Session(engine) as murfey_db:
        return (
            murfey_db.exec(select(MurfeySession).where(MurfeySession.id == session_id))
            .one()
            .visit
        )


async def submit_to_auth_endpoint(url_subpath: str, token: str) -> None:
    if auth_url:
        headers = (
            {}
            if security_config.auth_type == "cookie"
            else {"Authorization": f"Bearer {token}"}
        )
        cookies = (
            {security_config.cookie_key: token}
            if security_config.auth_type == "cookie"
            else {}
        )
        async with aiohttp.ClientSession(cookies=cookies) as session:
            async with session.get(
                f"{auth_url}/{url_subpath}",
                headers=headers,
            ) as response:
                success = response.status == 200
                validation_outcome: dict = await response.json()
        if not (success and validation_outcome.get("valid")):
            logger.warning("Unauthorised visit access request from frontend")
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="You do not have access to this visit",
                headers={"WWW-Authenticate": "Bearer"},
            )


async def validate_frontend_session_access(
    session_id: int,
    token: Annotated[str, Depends(oauth2_scheme)],
) -> int:
    """
    Validates whether a frontend request can access information about this session
    """
    visit_name = get_visit_name(session_id)
    await submit_to_auth_endpoint(f"validate_visit_access/{visit_name}", token)
    return session_id


async def validate_instrument_server_session_access(
    session_id: int,
    token: Annotated[str, Depends(instrument_oauth2_scheme)],
) -> int:
    """
    Validates whether an instrument request can access information about this session
    """
    visit_name = get_visit_name(session_id)

    if security_config.instrument_auth_url:
        async with aiohttp.ClientSession() as session:
            headers = (
                {}
                if not security_config.instrument_auth_type
                else {"Authorization": f"Bearer {token}"}
            )
            async with session.get(
                f"{security_config.instrument_auth_url}/validate_visit_access/{visit_name}",
                headers=headers,
            ) as response:
                success = response.status == 200
                validation_outcome = await response.json()
        if not (success and validation_outcome.get("valid")):
            logger.warning("Unauthorised visit access request from instrument")
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="You do not have access to this visit",
                headers={"WWW-Authenticate": "Bearer"},
            )
    return session_id


async def validate_user_instrument_access(
    instrument_name: str,
    token: Annotated[str, Depends(oauth2_scheme)],
) -> str:
    """
    Validates whether a frontend request can access information about this instrument
    """
    await submit_to_auth_endpoint(
        f"validate_instrument_access/{instrument_name}", token
    )
    return instrument_name


# Set validation conditions for the session ID based on where the request is from
MurfeySessionIDFrontend = Annotated[int, Depends(validate_frontend_session_access)]
MurfeySessionIDInstrument = Annotated[
    int, Depends(validate_instrument_server_session_access)
]

MurfeyInstrumentNameFrontend = Annotated[str, Depends(validate_user_instrument_access)]


"""
=======================================================================================
API ENDPOINTS AND HELPER FUNCTIONS/CLASSES
=======================================================================================
"""


def verify_password(plain_password: str, hashed_password: str) -> bool:
    return pwd_context.verify(plain_password, hashed_password)


def validate_user(username: str, password: str) -> bool:
    try:
        with Session(engine) as murfey_db:
            user = murfey_db.exec(select(User).where(User.username == username)).one()
    except Exception:
        return False
    return verify_password(password, user.hashed_password)


def create_access_token(data: dict, token: str = "") -> str:

    # If authenticating with password, auth URL needs a 'mint_session_token' endpoint
    if security_config.auth_type == "password":
        if auth_url and data.get("session"):
            session_id = data["session"]
            if not isinstance(session_id, int) and session_id > 0:
                # Check the session ID is alphanumeric for security
                raise ValueError("Session ID was invalid (not alphanumeric)")
            minted_token_response = requests.get(
                f"{auth_url}{url_path_for('auth.router', 'mint_session_token', session_id=session_id)}",
                headers={"Authorization": f"Bearer {token}"},
            )
            if minted_token_response.status_code != 200:
                raise RuntimeError(
                    f"Request received status code {minted_token_response.status_code} when trying to create session token"
                )
            return minted_token_response.json()["access_token"]

    to_encode = data.copy()

    # Make token for instrument
    encoded_jwt = jwt.encode(to_encode, SECRET_KEY, algorithm=ALGORITHM)
    return encoded_jwt


class Token(BaseModel):
    access_token: str
    token_type: str


@router.post("/token")
async def generate_token(
    form_data: Annotated[OAuth2PasswordRequestForm, Depends()],
) -> Token:
    # Only generate a token if it's a password
    if security_config.auth_type == "password":
        if auth_url:
            data = aiohttp.FormData()
            data.add_field("username", form_data.username)
            data.add_field("password", form_data.password)
            async with aiohttp.ClientSession() as session:
                async with session.post(
                    f"{auth_url}{url_path_for('auth.router', 'generate_token')}",
                    data=data,
                ) as response:
                    validated = response.status == 200
                    token = await response.json()
                    access_token = token.get("access_token")
        else:
            validated = validate_user(form_data.username, form_data.password)
            access_token = create_access_token(
                data={"user": form_data.username},
            )
        if not validated:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Incorrect username or password",
                headers={"WWW-Authenticate": "Bearer"},
            )
        return Token(access_token=access_token, token_type="bearer")

    # Return empty token otherwise
    return Token(access_token="", token_type="bearer")


@router.get("/sessions/{session_id}/token")
async def mint_session_token(session_id: MurfeySessionIDFrontend, db=murfey_db):
    visit = (
        db.exec(select(MurfeySession).where(MurfeySession.id == session_id)).one().visit
    )
    expiry_time = None
    if security_config.session_token_timeout:
        expiry_time = time.time() + security_config.session_token_timeout
    token = create_access_token(
        {
            "session": session_id,
            "visit": visit,
            "uuid": str(uuid4()),
            "expiry_time": expiry_time,
        }
    )
    return Token(access_token=token, token_type="bearer")


@router.get("/validate_token")
async def simple_token_validation(
    token: Annotated[str, Depends(validate_instrument_token)]
):
    return {"valid": True}



from logging import getLogger
from typing import List

import numpy as np
from sqlalchemy import func
from sqlmodel import Session, select

import murfey.server.prometheus as prom
from murfey.server import _transport_object
from murfey.server.feedback import (
    _app_id,
    _flush_class2d,
    _pj_id,
    _register_class_selection,
    _register_incomplete_2d_batch,
)
from murfey.util.config import get_machine_config
from murfey.util.db import (
    AutoProcProgram,
    CtfParameters,
    DataCollection,
    Movie,
    NotificationParameter,
    NotificationValue,
    ParticleSizes,
    ProcessingJob,
    SelectionStash,
)
from murfey.util.db import Session as MurfeySession
from murfey.util.db import SPAFeedbackParameters, SPARelionParameters
from murfey.util.processing_params import default_spa_parameters

logger = getLogger("murfey.workflows.spa.picking")


def _register_picked_particles_use_diameter(
    message: dict, _db: Session, demo: bool = False
):
    """Received picked particles from the autopick service"""
    # Add this message to the table of seen messages
    params_to_forward = message.get("extraction_parameters")
    assert isinstance(params_to_forward, dict)
    pj_id = _pj_id(message["program_id"], _db)
    ctf_params = CtfParameters(
        pj_id=pj_id,
        micrographs_file=params_to_forward["micrographs_file"],
        extract_file=params_to_forward["extract_file"],
        coord_list_file=params_to_forward["coord_list_file"],
        ctf_image=params_to_forward["ctf_values"]["CtfImage"],
        ctf_max_resolution=params_to_forward["ctf_values"]["CtfMaxResolution"],
        ctf_figure_of_merit=params_to_forward["ctf_values"]["CtfFigureOfMerit"],
        defocus_u=params_to_forward["ctf_values"]["DefocusU"],
        defocus_v=params_to_forward["ctf_values"]["DefocusV"],
        defocus_angle=params_to_forward["ctf_values"]["DefocusAngle"],
    )
    _db.add(ctf_params)
    _db.commit()
    _db.close()

    picking_db_len = _db.exec(
        select(func.count(ParticleSizes.id)).where(ParticleSizes.pj_id == pj_id)
    ).one()
    if picking_db_len > default_spa_parameters.nr_picks_before_diameter:
        # If there are enough particles to get a diameter
        instrument_name = (
            _db.exec(
                select(MurfeySession).where(MurfeySession.id == message["session_id"])
            )
            .one()
            .instrument_name
        )
        machine_config = get_machine_config(instrument_name=instrument_name)[
            instrument_name
        ]
        relion_params = _db.exec(
            select(SPARelionParameters).where(SPARelionParameters.pj_id == pj_id)
        ).one()
        relion_options = dict(relion_params)
        feedback_params = _db.exec(
            select(SPAFeedbackParameters).where(SPAFeedbackParameters.pj_id == pj_id)
        ).one()

        particle_diameter = relion_params.particle_diameter

        if feedback_params.picker_ispyb_id is None:
            if demo or not _transport_object:
                feedback_params.picker_ispyb_id = 1000
            else:
                assert feedback_params.picker_murfey_id is not None
                feedback_params.picker_ispyb_id = _transport_object.do_buffer_lookup(
                    message["program_id"], feedback_params.picker_murfey_id
                )
                if feedback_params.picker_ispyb_id is not None:
                    _flush_class2d(message["session_id"], message["program_id"], _db)
            _db.add(feedback_params)
            _db.commit()
            selection_stash = _db.exec(
                select(SelectionStash).where(SelectionStash.pj_id == pj_id)
            ).all()
            for s in selection_stash:
                _register_class_selection(
                    {
                        "session_id": s.session_id,
                        "class_selection_score": s.class_selection_score or 0,
                    },
                    _db=_db,
                    demo=demo,
                )
                _db.delete(s)
                _db.commit()

        if not particle_diameter:
            # If the diameter has not been calculated then find it
            picking_db = _db.exec(
                select(ParticleSizes.particle_size).where(ParticleSizes.pj_id == pj_id)
            ).all()
            particle_diameter = np.quantile(list(picking_db), 0.75)
            relion_params.particle_diameter = particle_diameter
            _db.add(relion_params)
            _db.commit()

            ctf_db = _db.exec(
                select(CtfParameters).where(CtfParameters.pj_id == pj_id)
            ).all()
            for saved_message in ctf_db:
                # Send on all saved messages to extraction
                _db.expunge(saved_message)
                zocalo_message: dict = {
                    "parameters": {
                        "micrographs_file": saved_message.micrographs_file,
                        "coord_list_file": saved_message.coord_list_file,
                        "output_file": saved_message.extract_file,
                        "pixel_size": (
                            relion_options["angpix"]
                            * relion_options["motion_corr_binning"]
                        ),
                        "ctf_image": saved_message.ctf_image,
                        "ctf_max_resolution": saved_message.ctf_max_resolution,
                        "ctf_figure_of_merit": saved_message.ctf_figure_of_merit,
                        "defocus_u": saved_message.defocus_u,
                        "defocus_v": saved_message.defocus_v,
                        "defocus_angle": saved_message.defocus_angle,
                        "particle_diameter": particle_diameter,
                        "downscale": relion_options["downscale"],
                        "kv": relion_options["voltage"],
                        "node_creator_queue": machine_config.node_creator_queue,
                        "session_id": message["session_id"],
                        "autoproc_program_id": _app_id(
                            _pj_id(message["program_id"], _db, recipe="em-spa-extract"),
                            _db,
                        ),
                        "batch_size": default_spa_parameters.batch_size_2d,
                    },
                    "recipes": ["em-spa-extract"],
                }
                if _transport_object:
                    zocalo_message["parameters"][
                        "feedback_queue"
                    ] = _transport_object.feedback_queue
                    _transport_object.send(
                        "processing_recipe", zocalo_message, new_connection=True
                    )
        else:
            # If the diameter is known then just send the new message
            particle_diameter = relion_params.particle_diameter
            zocalo_message = {
                "parameters": {
                    "micrographs_file": params_to_forward["micrographs_file"],
                    "coord_list_file": params_to_forward["coord_list_file"],
                    "output_file": params_to_forward["extract_file"],
                    "pixel_size": (
                        relion_options["angpix"] * relion_options["motion_corr_binning"]
                    ),
                    "ctf_image": params_to_forward["ctf_values"]["CtfImage"],
                    "ctf_max_resolution": params_to_forward["ctf_values"][
                        "CtfMaxResolution"
                    ],
                    "ctf_figure_of_merit": params_to_forward["ctf_values"][
                        "CtfFigureOfMerit"
                    ],
                    "defocus_u": params_to_forward["ctf_values"]["DefocusU"],
                    "defocus_v": params_to_forward["ctf_values"]["DefocusV"],
                    "defocus_angle": params_to_forward["ctf_values"]["DefocusAngle"],
                    "particle_diameter": particle_diameter,
                    "downscale": relion_options["downscale"],
                    "kv": relion_options["voltage"],
                    "node_creator_queue": machine_config.node_creator_queue,
                    "session_id": message["session_id"],
                    "autoproc_program_id": _app_id(
                        _pj_id(message["program_id"], _db, recipe="em-spa-extract"), _db
                    ),
                    "batch_size": default_spa_parameters.batch_size_2d,
                },
                "recipes": ["em-spa-extract"],
            }
            if _transport_object:
                zocalo_message["parameters"][
                    "feedback_queue"
                ] = _transport_object.feedback_queue
                _transport_object.send(
                    "processing_recipe", zocalo_message, new_connection=True
                )
            if demo:
                _register_incomplete_2d_batch(
                    {
                        "session_id": message["session_id"],
                        "program_id": message["program_id"],
                        "class2d_message": {
                            "particles_file": "Select/job009/particles_split_1.star",
                            "class2d_dir": "Class2D",
                            "batch_size": 50000,
                        },
                    },
                    _db=_db,
                    demo=demo,
                )

    else:
        # If not enough particles then save the new sizes
        particle_list = message.get("particle_diameters")
        assert isinstance(particle_list, list)
        for particle in particle_list:
            new_particle = ParticleSizes(pj_id=pj_id, particle_size=particle)
            _db.add(new_particle)
            _db.commit()
    _db.close()


def _register_picked_particles_use_boxsize(message: dict, _db: Session):
    """Received picked particles from the autopick service"""
    # Add this message to the table of seen messages
    params_to_forward = message.get("extraction_parameters")
    assert isinstance(params_to_forward, dict)

    instrument_name = (
        _db.exec(select(MurfeySession).where(MurfeySession.id == message["session_id"]))
        .one()
        .instrument_name
    )
    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]
    pj_id = _pj_id(message["program_id"], _db)
    ctf_params = CtfParameters(
        pj_id=pj_id,
        micrographs_file=params_to_forward["micrographs_file"],
        coord_list_file=params_to_forward["coord_list_file"],
        extract_file=params_to_forward["extract_file"],
        ctf_image=params_to_forward["ctf_values"]["CtfImage"],
        ctf_max_resolution=params_to_forward["ctf_values"]["CtfMaxResolution"],
        ctf_figure_of_merit=params_to_forward["ctf_values"]["CtfFigureOfMerit"],
        defocus_u=params_to_forward["ctf_values"]["DefocusU"],
        defocus_v=params_to_forward["ctf_values"]["DefocusV"],
        defocus_angle=params_to_forward["ctf_values"]["DefocusAngle"],
    )
    _db.add(ctf_params)
    _db.commit()
    _db.close()

    # Set particle diameter as zero and send box sizes
    relion_params = _db.exec(
        select(SPARelionParameters).where(SPARelionParameters.pj_id == pj_id)
    ).one()
    feedback_params = _db.exec(
        select(SPAFeedbackParameters).where(SPAFeedbackParameters.pj_id == pj_id)
    ).one()

    if feedback_params.picker_ispyb_id is None and _transport_object:
        assert feedback_params.picker_murfey_id is not None
        feedback_params.picker_ispyb_id = _transport_object.do_buffer_lookup(
            message["program_id"], feedback_params.picker_murfey_id
        )
        if feedback_params.picker_ispyb_id is not None:
            _flush_class2d(message["session_id"], message["program_id"], _db)
        _db.add(feedback_params)
        _db.commit()
        selection_stash = _db.exec(
            select(SelectionStash).where(SelectionStash.pj_id == pj_id)
        ).all()
        for s in selection_stash:
            _register_class_selection(
                {
                    "session_id": s.session_id,
                    "class_selection_score": s.class_selection_score or 0,
                },
                _db=_db,
            )
            _db.delete(s)
            _db.commit()

    # Send the message to extraction with the box sizes
    zocalo_message: dict = {
        "parameters": {
            "micrographs_file": params_to_forward["micrographs_file"],
            "coord_list_file": params_to_forward["coord_list_file"],
            "output_file": params_to_forward["extract_file"],
            "pixel_size": relion_params.angpix * relion_params.motion_corr_binning,
            "ctf_image": params_to_forward["ctf_values"]["CtfImage"],
            "ctf_max_resolution": params_to_forward["ctf_values"]["CtfMaxResolution"],
            "ctf_figure_of_merit": params_to_forward["ctf_values"]["CtfFigureOfMerit"],
            "defocus_u": params_to_forward["ctf_values"]["DefocusU"],
            "defocus_v": params_to_forward["ctf_values"]["DefocusV"],
            "defocus_angle": params_to_forward["ctf_values"]["DefocusAngle"],
            "particle_diameter": relion_params.particle_diameter,
            "boxsize": relion_params.boxsize,
            "small_boxsize": relion_params.small_boxsize,
            "downscale": relion_params.downscale,
            "kv": relion_params.voltage,
            "node_creator_queue": machine_config.node_creator_queue,
            "session_id": message["session_id"],
            "autoproc_program_id": _app_id(
                _pj_id(message["program_id"], _db, recipe="em-spa-extract"), _db
            ),
            "batch_size": default_spa_parameters.batch_size_2d,
        },
        "recipes": ["em-spa-extract"],
    }
    if _transport_object:
        zocalo_message["parameters"][
            "feedback_queue"
        ] = _transport_object.feedback_queue
        _transport_object.send("processing_recipe", zocalo_message, new_connection=True)
    _db.close()


def _request_email(
    failed_params: List[str], dcg_id: int, session_id: int, murfey_db: Session
) -> None:
    session = murfey_db.exec(
        select(MurfeySession).where(MurfeySession.id == session_id)
    ).one()
    config = get_machine_config(instrument_name=session.instrument_name)[
        session.instrument_name
    ]
    if _transport_object:
        _transport_object.send(
            config.notifications_queue,
            {
                "groupId": dcg_id,
                "message": f"The following parameters consistently exceeded the user set bounds: {failed_params}",
            },
            new_connection=True,
        )
        logger.debug(
            f"Sent notification to {config.notifications_queue!r} for "
            f"visit {session.visit!r}, data collection group ID {dcg_id} about the following abnormal parameters: \n"
            f"{', '.join([f'{p}' for p in failed_params])}"
        )
    return None


def _check_notifications(message: dict, murfey_db: Session) -> None:
    data_collection_hierarchy = murfey_db.exec(
        select(DataCollection, ProcessingJob, AutoProcProgram)
        .where(ProcessingJob.dc_id == DataCollection.id)
        .where(AutoProcProgram.pj_id == ProcessingJob.id)
        .where(AutoProcProgram.id == message["program_id"])
    ).all()
    dcgid = data_collection_hierarchy[0][0].dcg_id
    notification_parameters = murfey_db.exec(
        select(NotificationParameter).where(NotificationParameter.dcg_id == dcgid)
    ).all()
    failures = []
    for param in notification_parameters:
        if message.get(param.name) is not None:
            # Load instances of current parameter from database
            param_values = murfey_db.exec(
                select(NotificationValue).where(
                    NotificationValue.notification_parameter_id == param.id
                )
            ).all()
            param_values.sort(key=lambda x: x.index)

            # Drop oldest value if number of entries exceeds threshold
            param_value_to_drop = None
            if len(param_values) >= 25:
                param_value_to_drop = param_values[0]
                param_values = param_values[1:]

            # Add newest value to end of list
            param_values.append(
                NotificationValue(
                    notification_parameter_id=param.id,
                    index=param_values[-1].index + 1 if len(param_values) else 0,
                    within_bounds=param.min_value
                    <= message[param.name]
                    <= param.max_value,
                )
            )

            # Trigger message if this param has consistently exceeded the set threshold
            if (
                len(param_values) >= 25
                and sum(p.within_bounds for p in param_values) / len(param_values)
                < 0.25
            ):
                # If notifications disabled, enable them now
                trigger = False
                if not param.notification_active:
                    # Use a variable to trigger the notification for the first
                    # time within the first 500 messages received
                    if param_values[-1].index < 500:
                        logger.debug(
                            f"First abnormal instance of parameter {param.name!r} detected"
                        )
                        trigger = True
                    param.notification_active = True

                if param.num_instances_since_triggered >= 500 or trigger:
                    if not trigger:
                        logger.debug(
                            f"Parameter {param.name!r} has exceeded normal operating thresholds"
                        )
                    failures.append(param.name)
                    param.num_instances_since_triggered = 0
            else:
                # Only reset to False if there are more than 500 instances
                # to stop multiple triggers within the first 500
                if param.notification_active and param_values[-1].index > 500:
                    param.notification_active = False

            # Delete oldest value
            if param_value_to_drop is not None:
                murfey_db.delete(param_value_to_drop)

            # Add newest value and increment record of instances
            murfey_db.add(param_values[-1])
            param.num_instances_since_triggered += 1

    murfey_db.add_all(notification_parameters)
    murfey_db.commit()
    murfey_db.close()
    if failures:
        logger.debug(
            "Requested email notification for the following abnormal parameters: \n"
            f"{', '.join([f'{p}' for p in failures])}"
        )
        _request_email(failures, dcgid, message["session_id"], murfey_db)
    return None


def particles_picked(message: dict, murfey_db: Session) -> bool:
    movie = murfey_db.exec(
        select(Movie).where(Movie.murfey_id == message["motion_correction_id"])
    ).one()
    movie.preprocessed = True
    murfey_db.add(movie)
    murfey_db.commit()
    feedback_params = murfey_db.exec(
        select(SPAFeedbackParameters).where(
            SPAFeedbackParameters.pj_id == _pj_id(message["program_id"], murfey_db)
        )
    ).one()
    if feedback_params.estimate_particle_diameter:
        _register_picked_particles_use_diameter(message, murfey_db)
    else:
        _register_picked_particles_use_boxsize(message, murfey_db)
    prom.preprocessed_movies.labels(
        processing_job=_pj_id(message["program_id"], murfey_db)
    ).inc()
    _check_notifications(message, murfey_db)
    return True
